# MIT License
#
# Copyright (c) 2021 Dat Thanh Tran
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import torch
import torch.optim as optim
import pickle
from tqdm import tqdm
import os
import time
import numpy as np


def get_cosine_lr_scheduler(init_lr, final_lr):
    def lr_scheduler(n_epoch, epoch_idx):
        lr = final_lr + 0.5 * (init_lr - final_lr) * (1 + np.cos(np.pi * epoch_idx / n_epoch))
        return lr

    return lr_scheduler


def get_multiplicative_lr_scheduler(init_lr, drop_at, multiplicative_factor):
    def lr_scheduler(n_epoch, epoch_idx):
        lr = init_lr
        for epoch in drop_at:
            if epoch_idx + 1 >= epoch:
                lr *= multiplicative_factor
        return lr

    return lr_scheduler


class ClassifierTrainer:
    n_test_minibatch = 10

    def __init__(self,
                 n_epoch,
                 epoch_idx,
                 lr_scheduler,
                 optimizer,
                 weight_decay,
                 temp_dir,
                 checkpoint_freq=1,
                 print_freq=1,
                 use_progress_bar=False,
                 test_mode=False):

        assert epoch_idx < n_epoch,\
            'epoch_idx ("{}") must be lower than number of epochs ("{}")'.format(epoch_idx, n_epoch)

        self.n_epoch = n_epoch
        self.epoch_idx = epoch_idx
        self.lr_scheduler = lr_scheduler
        self.optimizer = optimizer
        self.weight_decay = weight_decay
        self.temp_dir = temp_dir
        self.checkpoint_freq = checkpoint_freq
        self.print_freq = print_freq
        self.use_progress_bar = use_progress_bar
        self.test_mode = test_mode

        self.metrics = ['cross_entropy', 'acc']
        self.monitor_metric = 'acc'
        self.monitor_direction = 'higher'

    def fit(self, model, train_loader, val_loader, test_loader, device, tensorboard_logger=None, logger_prefix=''):

        self.start_time = time.time()
        n_epoch_done = 0

        model.float()
        model.to(device)
        optimizer = self.get_optimizer(model)
        self.load_from_checkpoint(model, optimizer, device)

        while self.epoch_idx < self.n_epoch:
            # optimize one epoch
            self.optimize_epoch(model,
                                optimizer,
                                train_loader,
                                val_loader,
                                test_loader,
                                device)
            n_epoch_done += 1

            if self.print_freq > 0 and (self.epoch_idx + 1) % self.print_freq == 0:
                self.print_metrics(n_epoch_done)

            self.update_tensorboard(tensorboard_logger, logger_prefix)

            # save checkpoint
            if self.checkpoint_freq > 0 and (self.epoch_idx + 1) % self.checkpoint_freq == 0:
                checkpoint = {'epoch_idx': self.epoch_idx,
                              'model_state_dict': model.state_dict(),
                              'optimizer_state_dict': optimizer.state_dict(),
                              'metric_values': self.metric_values}

                checkpoint_file = os.path.join(self.temp_dir, 'checkpoint_{:09d}.pickle'.format(self.epoch_idx))
                fid = open(checkpoint_file, 'wb')
                pickle.dump(checkpoint, fid)
                fid.close()

            self.epoch_idx += 1

        # load the best model based on validation performance if exist, or train performance
        self.load_best(model)

        # return non-empty performance metrics
        performance = {}
        for metric in self.metric_values.keys():
            if len(self.metric_values[metric]) > 0:
                performance[metric] = self.metric_values[metric]

        return performance

    def load_best(self, model):
        # load the best model from checkpoints based on monitor_metric
        if len(self.metric_values['val_' + self.monitor_metric]) > 0:
            best_value = self.metric_values['val_' + self.monitor_metric][-1]
        else:
            best_value = self.metric_values['train_' + self.monitor_metric][-1]

        state_dict = model.state_dict()

        checkpoint_files = [os.path.join(self.temp_dir, f) for f in os.listdir(self.temp_dir)
                            if f.startswith('checkpoint_')]

        for filename in checkpoint_files:
            fid = open(filename, 'rb')
            checkpoint = pickle.load(fid)
            fid.close()

            if len(checkpoint['metric_values']['val_' + self.monitor_metric]) > 0:
                metric_value = checkpoint['metric_values']['val_' + self.monitor_metric][-1]
            else:
                metric_value = checkpoint['metric_values']['train_' + self.monitor_metric][-1]

            if (self.monitor_direction == 'lower' and metric_value < best_value) or\
                    (self.monitor_direction == 'higher' and metric_value > best_value):
                best_value = metric_value
                state_dict = checkpoint['model_state_dict']

        model.load_state_dict(state_dict)

    def get_optimizer(self, model):
        assert self.optimizer in ['adam', 'sgd'], 'Given optimizer "{}" is not supported'.format(self.optimizer)

        # get current learning rate
        lr = self.lr_scheduler(self.n_epoch, self.epoch_idx)

        # get separate batchnorm parameters and other parameters
        # if .get_parameters() is implemented in the model
        if hasattr(model, 'get_parameters') and callable(model.get_parameters):
            bn_params, other_params = model.get_parameters()

            if len(bn_params) > 0:
                params = [{'params': bn_params, 'weight_decay': 0},
                          {'params': other_params, 'weight_decay': self.weight_decay}]
            else:
                params = [{'params': other_params, 'weight_decay': self.weight_decay}]

            if self.optimizer == 'adam':
                optimizer = optim.Adam(params, lr=lr)
            else:
                optimizer = optim.SGD(params, lr=lr, momentum=0.9, nesterov=True)
        else:
            if self.optimizer == 'adam':
                optimizer = optim.Adam(model.parameters(), weight_decay=self.weight_decay, lr=lr)
            else:
                optimizer = optim.SGD(model.parameters(),
                                      weight_decay=self.weight_decay,
                                      lr=lr,
                                      momentum=0.9,
                                      nesterov=True)

        return optimizer

    def eval(self, model, loader, device):
        if loader is None:
            return {}

        model.eval()

        L = torch.nn.CrossEntropyLoss()
        n_correct = 0
        n_sample = 0
        loss = 0

        if self.test_mode:
            total_minibatch = min(self.n_test_minibatch, len(loader))
        else:
            total_minibatch = len(loader)

        with torch.no_grad():
            for minibatch_idx, (inputs, targets) in enumerate(loader):
                if minibatch_idx == total_minibatch:
                    break

                inputs = inputs.to(device)
                targets = targets.to(device).long().flatten()

                predictions = model(inputs)
                n_sample += inputs.size(0)
                loss += L(predictions, targets).item()
                n_correct += (predictions.argmax(dim=-1) == targets).sum().item()

        metrics = {'cross_entropy': loss / n_sample,
                   'acc': n_correct / n_sample}

        return metrics

    def update_lr(self,
                  optimizer):
        # update learning rate using lr_scheduler
        lr = self.lr_scheduler(self.n_epoch, self.epoch_idx)
        for param_group in optimizer.param_groups:
            param_group['lr'] = lr

    def update_loop(self, model, loader, optimizer, device):
        L = torch.nn.CrossEntropyLoss()

        if self.test_mode:
            total_minibatch = min(self.n_test_minibatch, len(loader))
        else:
            total_minibatch = len(loader)

        minibatch_idx = 0

        if self.use_progress_bar:
            loader = tqdm(loader, desc='#Epoch {}/{}: '.format(self.epoch_idx + 1, self.n_epoch), ncols=80, ascii=True)
        else:
            loader = loader

        for inputs, targets in loader:
            optimizer.zero_grad()
            self.update_lr(optimizer)

            inputs = inputs.to(device)
            targets = targets.to(device).long().flatten()

            predictions = model(inputs)
            loss = L(predictions, targets)
            loss.backward()
            optimizer.step()

            minibatch_idx += 1

            if minibatch_idx > total_minibatch:
                break

    def update_metrics(self, train_metrics, val_metrics, test_metrics):
        for metric in train_metrics.keys():
            if 'train_' + metric in self.metric_values.keys():
                self.metric_values['train_' + metric].append(train_metrics[metric])
            else:
                self.metric_values['train_' + metric] = [train_metrics[metric]]

        for metric in val_metrics.keys():
            if 'val_' + metric in self.metric_values.keys():
                self.metric_values['val_' + metric].append(val_metrics[metric])
            else:
                self.metric_values['val_' + metric] = [val_metrics[metric]]

        for metric in test_metrics.keys():
            if 'test_' + metric in self.metric_values.keys():
                self.metric_values['test_' + metric].append(test_metrics[metric])
            else:
                self.metric_values['test_' + metric] = [test_metrics[metric]]

    def print_metrics(self, n_epoch_done):
        start_time = self.start_time
        current_time = time.time()
        n_epoch_remain = self.n_epoch - n_epoch_done

        # compute the time taken
        time_taken = current_time - start_time
        hour_taken = int(time_taken / 3600)
        minute_taken = int((time_taken - hour_taken * 3600) / 60)
        second_taken = int((time_taken - hour_taken * 3600 - minute_taken * 60))

        # compute estimated time remain
        time_left = (time_taken / n_epoch_done) * n_epoch_remain
        hour_left = int(time_left / 3600)
        minute_left = int((time_left - hour_left * 3600) / 60)
        second_left = int((time_left - hour_left * 3600 - minute_left * 60))

        msg = '#Epoch {}/{}, '.format(self.epoch_idx + 1, self.n_epoch) +\
            'total time taken: {:d}:{:02d}:{:02d}, '.format(hour_taken, minute_taken, second_taken) +\
            'time remain: {:d}:{:02d}:{:02d}'.format(hour_left, minute_left, second_left)

        print(msg)

        names = list(self.metric_values.keys())
        names.sort()
        for name in names:
            if len(self.metric_values[name]) > 0:
                value = self.metric_values[name][-1]
                if isinstance(value, (int, float)):
                    print('--- {}: {:.6f}'.format(name, value))

    def update_tensorboard(self, tensorboard_logger, logger_prefix):
        names = list(self.metric_values.keys())
        names.sort()
        if tensorboard_logger is not None:
            for name in names:
                if len(self.metric_values[name]) > 0:
                    value = self.metric_values[name][-1]
                    if isinstance(value, (int, float)):
                        tensorboard_logger.add_scalar(tag='{}/{}'.format(logger_prefix, name),
                                                      scalar_value=value,
                                                      global_step=self.epoch_idx + 1)
                        tensorboard_logger.flush()

    def optimize_epoch(self,
                       model,
                       optimizer,
                       train_loader,
                       val_loader,
                       test_loader,
                       device):
        model.train()

        # perform parameter updates
        self.update_loop(model, train_loader, optimizer, device)

        # evaluate
        train_metrics = self.eval(model, train_loader, device)
        val_metrics = self.eval(model, val_loader, device)
        test_metrics = self.eval(model, test_loader, device)

        # append current performance to performance list
        self.update_metrics(train_metrics, val_metrics, test_metrics)

    def load_from_checkpoint(self, model, optimizer, device):
        if self.epoch_idx == -1:
            # load from latest checkpoint
            files = [os.path.join(self.temp_dir, f) for f in os.listdir(self.temp_dir)
                     if f.startswith('checkpoint_')]
            files.sort()

            if len(files) > 0:
                fid = open(files[-1], 'rb')
                checkpoint = pickle.load(fid)
                fid.close()
            else:
                checkpoint = None
        elif self.epoch_idx == 0:
            # train from scratch
            checkpoint = None
        else:
            # load specific checkpoint
            filename = os.path.join(self.temp_dir, 'checkpoint_{:9d}.pickle'.format(self.epoch_idx))
            assert os.path.exists(filename),\
                'checkpoint "{}" does not exist'.format(filename)

            fid = open(filename, 'rb')
            checkpoint = pickle.load(fid)
            fid.close()

        if checkpoint is None:
            self.epoch_idx = 0
            self.metric_values = {}
            prefixes = ['train', 'val', 'test']
            for prefix in prefixes:
                for m in self.metrics:
                    self.metric_values['{}_{}'.format(prefix, m)] = []
        else:
            # set the epoch index and previous metric values
            self.epoch_idx = checkpoint['epoch_idx'] + 1
            self.metric_values = checkpoint['metric_values']

            # load model state dict
            model.load_state_dict(checkpoint['model_state_dict'])

            # load optimizer state dict
            optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
            for state in optimizer.state.values():
                for k, v in state.items():
                    if torch.is_tensor(v):
                        state[k] = v.to(device)


class AutoRegressionTrainer(ClassifierTrainer):
    def __init__(self,
                 n_epoch,
                 epoch_idx,
                 lr_scheduler,
                 optimizer,
                 weight_decay,
                 temp_dir,
                 checkpoint_freq=1,
                 print_freq=1,
                 use_progress_bar=False,
                 test_mode=False):

        super(AutoRegressionTrainer, self).__init__(n_epoch, epoch_idx, lr_scheduler, optimizer, weight_decay,
                                                    temp_dir, checkpoint_freq, print_freq, use_progress_bar, test_mode)

        self.metrics = ['mean_squared_error']
        self.monitor_metric = 'mean_squared_error'
        self.monitor_direction = 'lower'

    def update_loop(self, model, loader, optimizer, device):
        L = torch.nn.MSELoss()

        if self.test_mode:
            total_minibatch = min(self.n_test_minibatch, len(loader))
        else:
            total_minibatch = len(loader)

        minibatch_idx = 0

        if self.use_progress_bar:
            loader = tqdm(loader, desc='#Epoch {}/{}: '.format(self.epoch_idx + 1, self.n_epoch), ncols=80, ascii=True)
        else:
            loader = loader

        for inputs, _ in loader:

            optimizer.zero_grad()
            self.update_lr(optimizer)

            inputs = inputs.to(device)

            predictions = model(inputs)
            loss = L(predictions, inputs)
            loss.backward()
            optimizer.step()

            minibatch_idx += 1
            if minibatch_idx > total_minibatch:
                break

    def eval(self, model, loader, device):
        if loader is None:
            return {}

        model.eval()

        L = torch.nn.MSELoss()
        n_sample = 0
        loss = 0

        if self.test_mode:
            total_minibatch = min(self.n_test_minibatch, len(loader))
        else:
            total_minibatch = len(loader)

        with torch.no_grad():
            for minibatch_idx, (inputs, _) in enumerate(loader):
                if minibatch_idx == total_minibatch:
                    break

                inputs = inputs.to(device)

                predictions = model(inputs)
                n_sample += inputs.size(0)
                loss += L(predictions, inputs).item()

        metrics = {'mean_squared_error': loss / n_sample}

        return metrics



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.nn as nn
import torchvision.models.densenet as densenet_models
import re

try:
    from torch.hub import load_state_dict_from_url
except ImportError:
    from torch.utils.model_zoo import load_url as load_state_dict_from_url

getters = {'densenet121': densenet_models.densenet121,
           'densenet161': densenet_models.densenet161,
           'densenet169': densenet_models.densenet169,
           'densenet201': densenet_models.densenet201,
           }


def extract_parameters_from_dense_block(block):
    bn_params, other_params = [], []
    for _, layer in block.items():
        bn_params.extend(list(layer.norm1.parameters()))
        bn_params.extend(list(layer.norm2.parameters()))

        other_params.extend(list(layer.conv1.parameters()))
        other_params.extend(list(layer.conv2.parameters()))

    return bn_params, other_params


def extract_parameters_from_transition_block(block):
    bn_params, other_params = [], []
    bn_params.extend(list(block.norm.parameters()))
    other_params.extend(list(block.conv.parameters()))

    return bn_params, other_params


class DenseNet(nn.Module):
    def __init__(self, model_name, n_class, pretrained):
        super(DenseNet, self).__init__()

        assert model_name in getters.keys()
        assert pretrained in ['', 'with_classifier', 'without_classifier']

        self.body = getters[model_name](pretrained=False, num_classes=n_class)

        if pretrained != '':
            # get pretrained weights
            state_dict = load_state_dict_from_url(densenet_models.model_urls[model_name], progress=False)

            # fix naming issue
            # as in https://github.com/pytorch/vision/blob/master/torchvision/models/densenet.py
            pattern = re.compile(
                r'^(.*denselayer\d+\.(?:norm|relu|conv))\.((?:[12])\.(?:weight|bias|running_mean|running_var))$')

            for key in list(state_dict.keys()):
                res = pattern.match(key)
                if res:
                    new_key = res.group(1) + res.group(2)
                    state_dict[new_key] = state_dict[key]
                    del state_dict[key]

            # remove the last classifier layer from pretrained weights
            if pretrained == 'without_classifier':
                cur_state_dict = self.body.state_dict()
                state_dict['classifier.weight'] = cur_state_dict['classifier.weight']
                state_dict['classifier.bias'] = cur_state_dict['classifier.bias']
            else:
                assert n_class == 1000,\
                    'For option pretrained="with_imagenet_classifier", ' +\
                    'the number of classes must be 1000. Provided value: {}'.format(n_class)

            self.body.load_state_dict(state_dict)

    def forward(self, x):
        return self.body(x)

    def get_parameters(self,):
        bn_params, other_params = [], []

        for layer in self.body.features:
            if isinstance(layer, densenet_models._DenseBlock):
                bn, other = extract_parameters_from_dense_block(layer)
                bn_params.extend(bn)
                other_params.extend(other)
            elif isinstance(layer, densenet_models._Transition):
                bn, other = extract_parameters_from_transition_block(layer)
                bn_params.extend(bn)
                other_params.extend(other)
            elif isinstance(layer, nn.BatchNorm2d):
                bn_params.extend(list(layer.parameters()))
            else:
                other_params.extend(list(layer.parameters()))

        other_params.extend(list(self.body.classifier.parameters()))

        return bn_params, other_params


class DenseNet121(DenseNet):
    def __init__(self, n_class, pretrained=False):
        super(DenseNet121, self).__init__('densenet121', n_class, pretrained)


class DenseNet161(DenseNet):
    def __init__(self, n_class, pretrained=False):
        super(DenseNet161, self).__init__('densenet161', n_class, pretrained)


class DenseNet169(DenseNet):
    def __init__(self, n_class, pretrained=False):
        super(DenseNet169, self).__init__('densenet169', n_class, pretrained)


class DenseNet201(DenseNet):
    def __init__(self, n_class, pretrained=False):
        super(DenseNet201, self).__init__('densenet201', n_class, pretrained)



# MIT License
#
# Copyright (c) 2017 liukuang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import torch.nn as nn
import torch.nn.functional as F
from . import model_utils


def conv3x3(in_planes, out_planes, stride=1):
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(in_planes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)

        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.ModuleList([
                nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes)]
            )
        else:
            self.shortcut = None

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))

        if self.shortcut is not None:
            for layer in self.shortcut:
                x = layer(x)

        out += x
        out = F.relu(out)
        return out

    def get_parameters(self,):
        bn_params = list(self.bn1.parameters()) + list(self.bn2.parameters())
        other_params = list(self.conv1.parameters()) + list(self.conv2.parameters())

        if self.shortcut is not None:
            bn_params.extend(list(self.shortcut[1].parameters()))
            other_params.extend(list(self.shortcut[0].parameters()))

        return bn_params, other_params


class PreActBlock(nn.Module):
    '''Pre-activation version of the BasicBlock.'''
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(PreActBlock, self).__init__()
        self.bn1 = nn.BatchNorm2d(in_planes)
        self.conv1 = conv3x3(in_planes, planes, stride)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv2 = conv3x3(planes, planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False)
            )

    def forward(self, x):
        out = F.relu(self.bn1(x))
        shortcut = self.shortcut(out)
        out = self.conv1(out)
        out = self.conv2(F.relu(self.bn2(out)))
        out += shortcut

        return out

    def get_parameters(self,):
        bn_params = list(self.bn1.parameters()) + list(self.bn2.parameters())
        other_params = list(self.conv1.parameters()) + list(self.conv2.parameters()) +\
            list(self.shortcut.parameters())

        return bn_params, other_params


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, in_planes, planes, stride=1):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, self.expansion*planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(self.expansion*planes)

        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.ModuleList([
                nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes)]
            )
        else:
            self.shortcut = None

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))

        if self.shortcut is not None:
            for layer in self.shortcut:
                x = layer(x)

        out += x
        out = F.relu(out)

        return out

    def get_parameters(self,):
        bn_params = list(self.bn1.parameters()) + list(self.bn2.parameters()) +\
            list(self.bn3.parameters())

        other_params = list(self.conv1.parameters()) + list(self.conv2.parameters()) +\
            list(self.conv3.parameters())

        if self.shortcut is not None:
            bn_params.extend(list(self.shortcut[1].parameters()))
            other_params.extend(list(self.shortcut[0].parameters()))

        return bn_params, other_params


class ResNet(nn.Module):
    def __init__(self, block, num_blocks, num_classes=10):
        super(ResNet, self).__init__()
        self.in_planes = 64

        self.conv1 = conv3x3(3, 64)
        self.bn1 = nn.BatchNorm2d(64)

        residual_layers = []

        residual_layers += self._make_layer(block, 64, num_blocks[0], stride=1)
        residual_layers += self._make_layer(block, 128, num_blocks[1], stride=2)
        residual_layers += self._make_layer(block, 256, num_blocks[2], stride=2)
        residual_layers += self._make_layer(block, 512, num_blocks[3], stride=2)

        self.residual_layers = nn.ModuleList(residual_layers)

        self.linear = nn.Linear(512 * block.expansion, num_classes)

        self.initialize()

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks-1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion

        return layers

    def forward(self, x):
        out = self.conv1(x)
        out = self.bn1(out)
        out = F.relu(out)

        for layer in self.residual_layers:
            out = layer(out)

        out = out.mean(-1).mean(-1)
        out = self.linear(out)

        return out

    def initialize(self,):
        for layer in self.modules():
            if isinstance(layer, nn.Conv2d):
                nn.init.kaiming_normal_(layer.weight)
            elif isinstance(layer, nn.BatchNorm2d):
                nn.init.constant_(layer.weight, 1)
                nn.init.constant_(layer.bias, 0.0)
            elif isinstance(layer, nn.Linear):
                nn.init.kaiming_normal_(layer.weight)
                nn.init.constant_(layer.bias, 0.0)

    def get_parameters(self,):
        bn_params, other_params = [], []

        other_params.extend(list(self.conv1.parameters()))
        bn_params.extend(list(self.bn1.parameters()))

        for layer in self.residual_layers:
            bn, other = layer.get_parameters()
            bn_params.extend(bn)
            other_params.extend(other)

        other_params.extend(list(self.linear.parameters()))

        return bn_params, other_params


def load_pretrained_weights(model, model_name, n_class, pretrained):
    if pretrained == 'with_classifier':
        if n_class == 10:
            model_name = 'cifar10_{}'.format(model_name)
        elif n_class == 100:
            model_name = 'cifar100_{}'.format(model_name)
        else:
            model_name = ''

        state_dict = model_utils.get_cifar_pretrained_weights(model_name)

        if state_dict is not None:
            model.load_state_dict(state_dict)

    elif pretrained == 'without_classifier':
        model_name = 'cifar100_{}'.format(model_name)
        state_dict = model_utils.get_cifar_pretrained_weights(model_name)
        if state_dict is not None:
            cur_state_dict = model.state_dict()
            state_dict['linear.weight'] = cur_state_dict['linear.weight']
            state_dict['linear.bias'] = cur_state_dict['linear.bias']
            model.load_state_dict(state_dict)


class ResNet18(ResNet):
    def __init__(self, n_class, pretrained):
        super(ResNet18, self).__init__(PreActBlock, [2, 2, 2, 2], num_classes=n_class)
        load_pretrained_weights(self, 'resnet18', n_class, pretrained)


class ResNet34(ResNet):
    def __init__(self, n_class, pretrained):
        super(ResNet34, self).__init__(BasicBlock, [3, 4, 6, 3], num_classes=n_class)
        load_pretrained_weights(self, 'resnet34', n_class, pretrained)


class ResNet50(ResNet):
    def __init__(self, n_class, pretrained):
        super(ResNet50, self).__init__(Bottleneck, [3, 4, 6, 3], num_classes=n_class)
        load_pretrained_weights(self, 'resnet50', n_class, pretrained)


class ResNet101(ResNet):
    def __init__(self, n_class, pretrained):
        super(ResNet101, self).__init__(Bottleneck, [3, 4, 6, 3], num_classes=n_class)
        load_pretrained_weights(self, 'resnet101', n_class, pretrained)


class ResNet152(ResNet):
    def __init__(self, n_class, pretrained):
        super(ResNet152, self).__init__(Bottleneck, [3, 8, 36, 3], num_classes=n_class)
        load_pretrained_weights(self, 'resnet152', n_class, pretrained)



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup
import os
from os.path import join
from configparser import ConfigParser
from setuptools import find_packages
from setuptools.command.install import install
import sys

author = 'OpenDR consortium'
author_email = 'tefas@csd.auth.gr'
url = 'https://github.com/opendr-eu/opendr'
license = 'LICENSE'
# Retrieve version
exec(open('src/opendr/_version.py').read())
try:
    __version__
except NameError:
    __version__ = '0.0'

# Read the long description
with open("description.txt") as f:
    long_description = f.read()

# Disable AVX2 for BCOLZ to ensure wider compatibility
os.environ['DISABLE_BCOLZ_AVX2'] = 'true'


def get_packages(module=None):
    packages = []
    if module:
        packages = find_packages(where="./src")
        module_short_name = module
        if module == 'engine':
            packages = [x for x in packages if 'engine' in x]
        else:
            module_short_name = module.split("/")[1]
            packages = [x for x in packages if module_short_name in x]
        name = "opendr-toolkit-" + module_short_name.replace("_", "-")
    else:
        name = "opendr-toolkit"

    packages.append('opendr.utils')
    packages.append('opendr.perception')
    packages.append('opendr.engine')
    packages.append('opendr.control')
    packages.append('opendr.planning')
    packages.append('opendr.simulation')
    packages.append('opendr')

    return name, packages


def generate_manifest(module=None):
    with open("MANIFEST.in", "w") as f:
        if module == "engine":
            f.write("recursive-include src/opendr/engine *\n")
            f.write("include src/opendr/engine *\n")
            f.write("include src/opendr/utils *\n")
        elif module:
            f.write("recursive-include " + join("src/opendr", module) + " *\n")
            f.write("include " + join("src/opendr", module.split("/")[0]) + " *\n")

        f.write("exclude src/opendr/__init__.py \n")
        f.write("include description.txt \n")
        f.write("include packages.txt \n")
        f.write("include README.md \n")
        f.write("include src/opendr/_version.py \n")
        f.write("include src/opendr/_setup.py \n")


def get_description(module=None):
    if module:
        return 'Open Deep Learning Toolkit for Robotics (submodule: ' + module + ')'
    else:
        return 'Open Deep Learning Toolkit for Robotics'


def get_dependencies(current_module):
    dependencies = []
    skipped_dependencies = []
    post_install = []
    # Read all the dependencies.ini for each tool category
    if current_module:
        # Get all subfolders
        paths = ['.']

        for file in os.listdir(join("src/opendr", current_module)):
            if os.path.isdir(join("src/opendr", current_module, file)):
                paths.append(file)

        for path in paths:
            try:
                parser = ConfigParser()
                parser.read(join("src/opendr", current_module, path, 'dependencies.ini'))
                try:
                    runtime_deps = parser.get("runtime", "python").split('\n')
                except Exception:
                    runtime_deps = []
                try:
                    compilation_deps = parser.get("compilation", "python").split('\n')
                except Exception:
                    compilation_deps = []
                try:
                    opendr_deps = parser.get("runtime", "opendr").split('\n')
                except Exception:
                    opendr_deps = []
                try:
                    scripts = parser.get("runtime", "post-install").split('\n')
                    for x in scripts:
                        post_install.append(x)
                except Exception:
                    pass

            except Exception:
                pass

            deps = [x for x in list(set(runtime_deps + compilation_deps)) if x != '']
            # Add dependencies found (filter git-based ones and local ones)
            for x in deps:
                if 'git' in x or '${OPENDR_HOME}' in x:
                    skipped_dependencies.append(x)
                else:
                    dependencies.append(x)
            for x in opendr_deps:
                dependencies.append(x)

        dependencies = list(set(dependencies))
        skipped_dependencies = list(set(skipped_dependencies))
        post_install = list(set(post_install))
    else:
        with open("packages.txt", "r") as f:
            packages = [x.strip() for x in f.readlines()]
        for package in packages:
            if '/' in package:
                dependencies.append('opendr-toolkit-' + package.split('/')[1].replace('_', '-'))
            elif package != 'opendr':
                dependencies.append('opendr-toolkit-' + package.replace('_', '-'))

    return dependencies, skipped_dependencies, post_install


def get_data_files(module):
    data_files = []
    if module:
        for root, dirs, files in os.walk(join("src", "opendr", module)):
            for file in files:
                file_extension = file.split(".")[-1]
                # Add all files except from shared libraries
                if file_extension != "so" and file_extension != "py":
                    data_files.append(join(root.replace("src/opendr/", ""), file))
    return data_files


def build_package(module):
    if module == "opendr":
        # Flag to enable building opendr-metapackage
        module = None

    if module == 'perception/object_detection_2d':
        from Cython.Build import cythonize
        import numpy
        extra_params = {
            'ext_modules':
                cythonize([join("src/opendr/perception/object_detection_2d/retinaface/algorithm/cython/*.pyx")]),
            'include_dirs': [numpy.get_include()]}
    else:
        extra_params = {}

    name, packages = get_packages(module)
    dependencies, skipped_dependencies, post_install = get_dependencies(module)
    generate_manifest(module)

    # Define class for post installation scripts
    class PostInstallScripts(install):
        def run(self):
            install.run(self)
            import subprocess

            # Install potential git and local repos during post installation
            for package in skipped_dependencies:
                if 'git' in package:
                    subprocess.call([sys.executable, '-m', 'pip', 'install', package])
                if '${OPENDR_HOME}' in package:
                    subprocess.call([sys.executable, '-m', 'pip', 'install', package.replace('${OPENDR_HOME}', '.')])

            if post_install:
                for cmd in post_install:
                    print("Running ", cmd)
                    subprocess.call(cmd.split(' '))

    setup(
        name=name,
        version=__version__,
        description=get_description(module),
        long_description=long_description,
        author=author,
        author_email=author_email,
        packages=packages,
        url=url,
        license=license,
        package_dir={"": "./src"},
        install_requires=dependencies,
        cmdclass={
            'develop': PostInstallScripts,
            'install': PostInstallScripts,
        },
        package_data={'': get_data_files(module)},
        **extra_params
    )



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
import os
import numpy as np
from opendr.engine.data import PointCloud


class Dataset(ABC):
    """
    Dataset abstract class is used to provide a common ancestor type for dataset classes.
    All classes that provide dataset-related functionality must inherit this class.
    """

    def __init__(self):
        pass


class DatasetIterator(Dataset):
    """
    DatasetIterator serves as an abstraction layer over the different types of datasets.
    In this way it provides the opportunity to users to implement different kinds of datasets, while
    providing a uniform interface.

    This class provides the following abstract methods:
    - __getitem__(i), a getter that allows for retrieving the i-th sample of the dataset, along with its annotation
    - __len__(), which allows for getting the size of the dataset
    """
    def __init__(self):
        super().__init__()

    @abstractmethod
    def __getitem__(self, idx):
        """
        This method is used for loading the idx-th sample of a dataset along with its annotation.

        :param idx: the index of the sample to load
        :return: the idx-th sample and its annotation
        :rtype: Tuple of (Data, Target)
        """
        pass

    @abstractmethod
    def __len__(self):
        """
        This method returns the size of the dataset.

        :return: the size of the dataset
        :rtype: int
        """
        pass


class MappedDatasetIterator(DatasetIterator):
    """
    MappedDatasetIterator allows to transform elements of the original DatasetIterator.

    This class provides the following methods:
    - __getitem__(i), a getter that allows for retrieving the i-th sample of the dataset, along with its annotation
    - __len__(), which allows for getting the size of the dataset
    """
    def __init__(self, original, map_function):
        super().__init__()
        self.map_function = map_function
        self.original = original

    def __getitem__(self, idx):
        """
        This method is used for loading the idx-th sample of a dataset along with its annotation.

        :param idx: the index of the sample to load
        :return: the idx-th sample and its annotation
        :rtype: Tuple of (Data, Target)
        """

        return self.map_function(self.original[idx])

    def __len__(self):
        """
        This method returns the size of the dataset.

        :return: the size of the dataset
        :rtype: int
        """

        return len(self.original)


class ExternalDataset(Dataset):
    """
    ExternalDataset provides a way for handling well-known external dataset formats
    (e.g., COCO, PascalVOC, Imagenet, etc.) directly by OpenDR, without requiring any special effort
    from the users for writing a specific loader.

    Class attributes:
    - path: path to the external dataset
    - dataset_type:	type of the dataset format (e.g., “voc”, “coco”, “imagenet”, etc.)
    """
    def __init__(self, path, dataset_type):
        super().__init__()
        self.path = path
        self.dataset_type = dataset_type

    @property
    def path(self):
        """
        Getter of path field.
        This returns the path.

        :return: path to the external dataset
        :rtype: str
        """
        return self._path

    @path.setter
    def path(self, value):
        """
        Setter of path field. This will perform the necessary type checking.

        :param value: new path value
        :type value: str
        """
        if type(value) != str:
            raise ValueError('path should be a str')
        else:
            self._path = value

    @property
    def dataset_type(self):
        """
        Getter of dataset_type field.
        This returns the dataset_type.

        :return: type of the dataset format (e.g., “voc”, “coco”, “imagenet”, etc.)
        :rtype: str
        """
        return self._dataset_type

    @dataset_type.setter
    def dataset_type(self, value):
        """
        Setter of dataset_type field. This will perform the necessary type checking.

        :param value: new dataset_type value
        :type value: str
        """
        if type(value) != str:
            raise ValueError('dataset_type should be a str')
        else:
            self._dataset_type = value


class PointCloudsDatasetIterator(DatasetIterator):
    def __init__(self, path, num_point_features=4):
        super().__init__()

        self.path = path
        self.num_point_features = num_point_features
        self.files = sorted(os.listdir(path))

    def __getitem__(self, idx):
        data = np.fromfile(
            str(self.path + "/" + self.files[idx]), dtype=np.float32, count=-1
        ).reshape([-1, self.num_point_features])

        return PointCloud(data)

    def __len__(self):
        return len(self.files)



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.nn as nn
import torchvision.models.resnet as resnet_models

try:
    from torch.hub import load_state_dict_from_url
except ImportError:
    from torch.utils.model_zoo import load_url as load_state_dict_from_url

getters = {'resnet18': resnet_models.resnet18,
           'resnet34': resnet_models.resnet34,
           'resnet50': resnet_models.resnet50,
           'resnet101': resnet_models.resnet101,
           'resnet152': resnet_models.resnet152,
           'resnext50_32x4d': resnet_models.resnext50_32x4d,
           'resnext101_32x8d': resnet_models.resnext101_32x8d,
           'wide_resnet50_2': resnet_models.wide_resnet50_2,
           'wide_resnet101_2': resnet_models.wide_resnet101_2,
           }


class ResNet(nn.Module):
    def __init__(self, model_name, n_class, pretrained):
        super(ResNet, self).__init__()

        assert model_name in getters.keys()
        assert pretrained in ['', 'with_classifier', 'without_classifier']

        self.body = getters[model_name](pretrained=False, num_classes=n_class)

        if pretrained != '':
            # get pretrained weights
            state_dict = load_state_dict_from_url(resnet_models.model_urls[model_name], progress=False)

            # remove the last classifier layer from pretrained weights
            if pretrained == 'without_classifier':
                cur_state_dict = self.body.state_dict()
                state_dict['fc.weight'] = cur_state_dict['fc.weight']
                state_dict['fc.bias'] = cur_state_dict['fc.bias']
            else:
                assert n_class == 1000,\
                    'For option pretrained="with_imagenet_classifier", ' +\
                    'the number of classes must be 1000. Provided value: {}'.format(n_class)

            self.body.load_state_dict(state_dict)

    def forward(self, x):
        return self.body(x)

    def get_parameters(self,):
        bn_params, other_params = [], []

        other_params.extend(list(self.body.conv1.parameters()))
        bn_params.extend(list(self.body.bn1.parameters()))

        for block in [self.body.layer1, self.body.layer2, self.body.layer3, self.body.layer4]:
            for layer in block:
                if isinstance(layer, nn.BatchNorm2d):
                    bn_params.extend(list(layer.parameters()))
                else:
                    other_params.extend(list(layer.parameters()))

        other_params.extend(list(self.body.fc.parameters()))

        return bn_params, other_params


class ResNet18(ResNet):
    def __init__(self, n_class, pretrained=False):
        super(ResNet18, self).__init__('resnet18', n_class, pretrained)


class ResNet34(ResNet):
    def __init__(self, n_class, pretrained=False):
        super(ResNet34, self).__init__('resnet34', n_class, pretrained)


class ResNet50(ResNet):
    def __init__(self, n_class, pretrained=False):
        super(ResNet50, self).__init__('resnet50', n_class, pretrained)


class ResNet101(ResNet):
    def __init__(self, n_class, pretrained=False):
        super(ResNet101, self).__init__('resnet101', n_class, pretrained)


class ResNet152(ResNet):
    def __init__(self, n_class, pretrained=False):
        super(ResNet152, self).__init__('resnet152', n_class, pretrained)


class ResNext50_32x4d(ResNet):
    def __init__(self, n_class, pretrained=False):
        super(ResNext50_32x4d, self).__init__('resnext50_32x4d', n_class, pretrained)


class ResNext101_32x8d(ResNet):
    def __init__(self, n_class, pretrained=False):
        super(ResNext101_32x8d, self).__init__('resnext101_32x8d', n_class, pretrained)


class WideResNet50_2(ResNet):
    def __init__(self, n_class, pretrained=False):
        super(WideResNet50_2, self).__init__('wide_resnet50_2', n_class, pretrained)


class WideResNet101_2(ResNet):
    def __init__(self, n_class, pretrained=False):
        super(WideResNet101_2, self).__init__('wide_resnet101_2', n_class, pretrained)



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import backbones
import torch
import torch.nn as nn
import torch.nn.functional as F


def get_builtin_backbones():
    return list(backbones.models.keys())


class MultilinearMapping(nn.Module):
    def __init__(self, input_shape, output_shape):
        super(MultilinearMapping, self).__init__()

        assert len(input_shape) == 3,\
            'Input shape must be a 3-tuple containing (#row, #col, #channel). Received "{}"'.format(input_shape)

        assert len(output_shape) == 3,\
            'Output shape must be a 3-tuple containing (#row, #col, #channel). Received "{}"'.format(output_shape)

        # reorder input shape and compressed shape to (#channel, #row, #col)
        input_shape = (input_shape[2], input_shape[0], input_shape[1])
        output_shape = (output_shape[2], output_shape[0], output_shape[1])

        # parameters to map the channel dimension
        use_first_mode = False if (input_shape[0] == 1 and output_shape[0] == 1) else True
        if use_first_mode:
            self.W1 = nn.Parameter(data=torch.Tensor(output_shape[0], input_shape[0]),
                                   requires_grad=True)
        else:
            self.W1 = None

        # parameters to map the spatial dimension (rows)
        self.W2 = nn.Parameter(data=torch.Tensor(output_shape[1], input_shape[1]),
                               requires_grad=True)

        # parameters to map the spatial dimension (cols)
        self.W3 = nn.Parameter(data=torch.Tensor(output_shape[2], input_shape[2]),
                               requires_grad=True)

        # initialization
        nn.init.kaiming_uniform_(self.W3)
        nn.init.kaiming_uniform_(self.W2)
        if self.W1 is not None:
            nn.init.kaiming_uniform_(self.W1)

    def forward(self, x):
        x = self.nmodeproduct(x, self.W2, 2)
        x = self.nmodeproduct(x, self.W3, 3)
        if self.W1 is not None:
            x = self.nmodeproduct(x, self.W1, 1)

        return x

    def nmodeproduct(self, x, W, mode):
        if mode == 1:
            y = torch.transpose(x, 1, 3)
            y = F.linear(y, W)
            y = torch.transpose(y, 1, 3)
        elif mode == 2:
            y = torch.transpose(x, 2, 3)
            y = F.linear(y, W)
            y = torch.transpose(y, 2, 3)
        else:
            y = F.linear(x, W)

        return y


class SenseSynth(nn.Module):
    def __init__(self, input_shape, compressed_shape):
        super(SenseSynth, self).__init__()

        # sensing module
        self.sensing_module = MultilinearMapping(input_shape, compressed_shape)

        # synthesis module
        self.synthesis_module = MultilinearMapping(compressed_shape, input_shape)

    def forward(self, x):
        x = self.sensing_module(x)
        x = self.synthesis_module(x)
        return x


class CompressiveLearner(nn.Module):
    def __init__(self, input_shape, compressed_shape, n_class, backbone, pretrained_backbone):

        super(CompressiveLearner, self).__init__()

        if isinstance(backbone, str):
            # built-in backbone given as a string
            assert backbone in backbones.models.keys(),\
                'Given `backbone` parameter "{}" is not supported. '.format(backbone) +\
                'Supported `backbone` includes:\n' +\
                '\n'.join(['"{}"'.format(name) for name in backbones.models.keys()])

            self.backbone = backbones.models[backbone](n_class, pretrained_backbone)

        else:
            assert isinstance(backbone, torch.nn.Module),\
                '`backbone` parameter must be a string indicating built-in backbone ' +\
                'or an instance of torch.nn.Module\n' +\
                'received backbone of type: {}'.format(type(backbone))

            self.backbone = backbone

        self.sense_synth_module = SenseSynth(input_shape, compressed_shape)

    def forward(self, x):
        x = self.sense_synth_module(x)
        x = self.backbone(x)
        return x

    def infer_from_measurement(self, x):
        x = self.sense_synth_module.synthesis_module(x)
        x = self.backbone(x)
        return x

    def get_parameters(self,):
        bn_params, other_params = [], []
        if hasattr(self.backbone, 'get_parameters') and callable(self.backbone.get_parameters):
            backbone_bn_params, backbone_other_params = self.backbone.get_parameters()
            bn_params.extend(backbone_bn_params)
            other_params.extend(backbone_other_params)
            other_params.extend(list(self.sense_synth_module.parameters()))
        else:
            other_params.extend(list(self.sense_synth_module.parameters()))
            other_params.extend(list(self.backbone.parameters()))

        return bn_params, other_params



from opendr.perception.compressive_learning.multilinear_compressive_learning.multilinear_compressive_learner import (
    MultilinearCompressiveLearner, get_builtin_backbones, PRETRAINED_COMPRESSED_SHAPE)

__all__ = ['MultilinearCompressiveLearner', 'get_builtin_backbones', 'PRETRAINED_COMPRESSED_SHAPE']



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pickle
from urllib.request import urlretrieve


def get_cifar_pretrained_weights(model_name):
    if model_name == '':
        return

    home_dir = os.path.expanduser('~')
    model_dir = os.path.join(home_dir,
                             '.cache',
                             'opendr',
                             'checkpoints',
                             'perception',
                             'compressive_learning',
                             'backbone')

    if not os.path.exists(model_dir):
        os.makedirs(model_dir, exist_ok=True)

    model_file = os.path.join(model_dir, '{}.pickle'.format(model_name))

    if not os.path.exists(model_file):
        server_url = 'ftp://opendrdata.csd.auth.gr/perception/compressive_learning/backbone/'
        model_url = os.path.join(server_url, '{}.pickle'.format(model_name))
        urlretrieve(model_url, model_file)
        print('Pretrained backbone model downloaded')

    fid = open(model_file, 'rb')
    state_dict = pickle.load(fid)['state_dict']
    fid.close()

    return state_dict



# MIT License
#
# Copyright (c) 2017 liukuang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import torch.nn as nn
from . import model_utils

cfg = {
    'VGG11': [64, 'M', 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'VGG13': [64, 64, 'M', 128, 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'VGG16': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M', 512, 512, 512, 'M', 512, 512, 512, 'M'],
    'VGG19': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512, 512, 'M', 512, 512, 512, 512, 'M'],
}


class VGG(nn.Module):
    def __init__(self, vgg_name, n_class):
        super(VGG, self).__init__()
        self.features = nn.ModuleList(self._make_layers(cfg[vgg_name]))
        self.classifier = nn.Linear(512, n_class)
        self.initialize()

    def forward(self, x):
        for layer in self.features:
            x = layer(x)

        x = x.view(x.size(0), -1)
        x = self.classifier(x)

        return x

    def _make_layers(self, cfg):
        layers = []
        in_channels = 3
        for x in cfg:
            if x == 'M':
                layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
            else:
                layers += [nn.Conv2d(in_channels, x, kernel_size=3, padding=1),
                           nn.BatchNorm2d(x),
                           nn.ReLU(inplace=True)]
                in_channels = x
        layers += [nn.AvgPool2d(kernel_size=1, stride=1)]

        return layers

    def initialize(self,):
        for layer in self.modules():
            if isinstance(layer, nn.Conv2d):
                nn.init.kaiming_normal_(layer.weight)
                nn.init.constant_(layer.bias, 0.0)
            elif isinstance(layer, nn.BatchNorm2d):
                nn.init.constant_(layer.weight, 1)
                nn.init.constant_(layer.bias, 0.0)
            elif isinstance(layer, nn.Linear):
                nn.init.kaiming_normal_(layer.weight)
                nn.init.constant_(layer.bias, 0.0)

    def get_parameters(self,):
        bn_params, other_params = [], []
        for layer in self.features:
            if isinstance(layer, nn.BatchNorm2d):
                bn_params.extend(list(layer.parameters()))
            else:
                other_params.extend(list(layer.parameters()))

        other_params.extend(list(self.classifier.parameters()))

        return bn_params, other_params


def load_pretrained_weights(model, model_name, n_class, pretrained):
    if pretrained == 'with_classifier':
        if n_class == 10:
            model_name = 'cifar10_{}'.format(model_name)
        elif n_class == 100:
            model_name = 'cifar100_{}'.format(model_name)
        else:
            model_name = ''

        state_dict = model_utils.get_cifar_pretrained_weights(model_name)

        if state_dict is not None:
            model.load_state_dict(state_dict)

    elif pretrained == 'without_classifier':
        model_name = 'cifar100_{}'.format(model_name)
        state_dict = model_utils.get_cifar_pretrained_weights(model_name)

        if state_dict is not None:
            cur_state_dict = model.state_dict()
            state_dict['classifier.weight'] = cur_state_dict['classifier.weight']
            state_dict['classifier.bias'] = cur_state_dict['classifier.bias']
            model.load_state_dict(state_dict)


class VGG11(VGG):
    def __init__(self, n_class, pretrained):
        super(VGG11, self).__init__('VGG11', n_class)
        load_pretrained_weights(self, 'vgg11', n_class, pretrained)


class VGG13(VGG):
    def __init__(self, n_class, pretrained):
        super(VGG13, self).__init__('VGG13', n_class)
        load_pretrained_weights(self, 'vgg13', n_class, pretrained)


class VGG16(VGG):
    def __init__(self, n_class, pretrained):
        super(VGG16, self).__init__('VGG16', n_class)
        load_pretrained_weights(self, 'vgg16', n_class, pretrained)


class VGG19(VGG):
    def __init__(self, n_class, pretrained):
        super(VGG19, self).__init__('VGG19', n_class)
        load_pretrained_weights(self, 'vgg19', n_class, pretrained)



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.nn as nn
import torchvision.models.vgg as vgg_models

try:
    from torch.hub import load_state_dict_from_url
except ImportError:
    from torch.utils.model_zoo import load_url as load_state_dict_from_url

getters = {'vgg11': vgg_models.vgg11,
           'vgg11_bn': vgg_models.vgg11_bn,
           'vgg13': vgg_models.vgg13,
           'vgg13_bn': vgg_models.vgg13_bn,
           'vgg16': vgg_models.vgg16,
           'vgg16_bn': vgg_models.vgg16_bn,
           'vgg19': vgg_models.vgg19,
           'vgg19_bn': vgg_models.vgg19_bn}


class VGG(nn.Module):
    def __init__(self, model_name, n_class, pretrained):
        super(VGG, self).__init__()

        assert model_name in getters.keys()
        assert pretrained in ['', 'with_classifier', 'without_classifier']

        self.body = getters[model_name](pretrained=False, num_classes=n_class)

        if pretrained != '':
            # get pretrained weights
            state_dict = load_state_dict_from_url(vgg_models.model_urls[model_name], progress=False)

            # remove the last classifier layer from pretrained weights
            if pretrained == 'without_classifier':
                cur_state_dict = self.body.state_dict()
                state_dict['classifier.6.weight'] = cur_state_dict['classifier.6.weight']
                state_dict['classifier.6.bias'] = cur_state_dict['classifier.6.bias']
            else:
                assert n_class == 1000,\
                    'For option pretrained="with_imagenet_classifier", ' +\
                    'the number of classes must be 1000. Provided value: {}'.format(n_class)

            self.body.load_state_dict(state_dict)

    def forward(self, x):
        return self.body(x)

    def get_parameters(self,):
        bn_params, other_params = [], []
        for layer in self.body.features:
            if isinstance(layer, nn.BatchNorm2d):
                bn_params.extend(list(layer.parameters()))
            else:
                other_params.extend(list(layer.parameters()))

        other_params.extend(list(self.body.classifier.parameters()))

        return bn_params, other_params


class VGG11(VGG):
    def __init__(self, n_class, pretrained=False):
        super(VGG11, self).__init__('vgg11', n_class, pretrained)


class VGG11_BN(VGG):
    def __init__(self, n_class, pretrained=False):
        super(VGG11_BN, self).__init__('vgg11_bn', n_class, pretrained)


class VGG13(VGG):
    def __init__(self, n_class, pretrained=False):
        super(VGG13, self).__init__('vgg13', n_class, pretrained)


class VGG13_BN(VGG):
    def __init__(self, n_class, pretrained=False):
        super(VGG13_BN, self).__init__('vgg13_bn', n_class, pretrained)


class VGG16(VGG):
    def __init__(self, n_class, pretrained=False):
        super(VGG16, self).__init__('vgg16', n_class, pretrained)


class VGG16_BN(VGG):
    def __init__(self, n_class, pretrained=False):
        super(VGG16_BN, self).__init__('vgg16_bn', n_class, pretrained)


class VGG19(VGG):
    def __init__(self, n_class, pretrained=False):
        super(VGG19, self).__init__('vgg19', n_class, pretrained)


class VGG19_BN(VGG):
    def __init__(self, n_class, pretrained=False):
        super(VGG19_BN, self).__init__('vgg19_bn', n_class, pretrained)



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch.nn as nn
from . import model_utils


class AllCNN(nn.Module):
    def __init__(self, n_class, pretrained):
        super(AllCNN, self).__init__()

        self.block1 = nn.ModuleList([
            nn.Conv2d(in_channels=3, out_channels=36, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(36),
            nn.ReLU(),
            nn.Conv2d(in_channels=36, out_channels=96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96),
            nn.ReLU(),
            nn.Conv2d(in_channels=96, out_channels=96, kernel_size=3, stride=2, padding=1),
            nn.BatchNorm2d(96),
            nn.ReLU(),
            nn.Dropout(0.2),
        ])

        self.block2 = nn.ModuleList([
            nn.Conv2d(in_channels=96, out_channels=192, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(192),
            nn.ReLU(),
            nn.Conv2d(in_channels=192, out_channels=192, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(192),
            nn.ReLU(),
            nn.Conv2d(in_channels=192, out_channels=192, kernel_size=3, stride=2, padding=1),
            nn.BatchNorm2d(192),
            nn.ReLU(),
            nn.Dropout(0.2),
        ])

        self.block3 = nn.ModuleList([
            nn.Conv2d(in_channels=192, out_channels=192, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(192),
            nn.ReLU(),
            nn.Conv2d(in_channels=192, out_channels=192, kernel_size=1, stride=1, padding=0),
            nn.BatchNorm2d(192),
            nn.ReLU(),
        ])

        self.classifier = nn.ModuleList([
            nn.Conv2d(in_channels=192, out_channels=n_class, kernel_size=1, stride=1, padding=0),
            nn.BatchNorm2d(n_class),
            nn.ReLU(),
        ])

        if pretrained == 'with_classifier':
            if n_class == 10:
                model_name = 'cifar10_allcnn'
            elif n_class == 100:
                model_name = 'cifar100_allcnn'
            else:
                model_name = ''

            state_dict = model_utils.get_cifar_pretrained_weights(model_name)
            if state_dict is not None:
                self.load_state_dict(state_dict)
        elif pretrained == 'without_classifier':
            model_name = 'cifar100_allcnn'
            state_dict = model_utils.get_cifar_pretrained_weights(model_name)
            cur_state_dict = self.state_dict()
            for key in cur_state_dict.keys():
                if key.startswith('classifier'):
                    state_dict[key] = cur_state_dict[key]

    def forward(self, x):
        for layer in self.block1:
            x = layer(x)
        for layer in self.block2:
            x = layer(x)
        for layer in self.block3:
            x = layer(x)
        for layer in self.classifier:
            x = layer(x)
        x = x.mean(dim=-1).mean(dim=-1)
        return x

    def initialize(self,):
        for layer in self.modules():
            if isinstance(layer, nn.Conv2d):
                nn.init.kaiming_normal_(layer.weight)
                nn.init.constant_(layer.bias, 0.0)
            elif isinstance(layer, nn.BatchNorm2d):
                nn.init.constant_(layer.weight, 1)
                nn.init.constant_(layer.bias, 0.0)
            elif isinstance(layer, nn.Linear):
                nn.init.kaiming_normal_(layer.weight)
                nn.init.constant_(layer.bias, 0.0)

    def get_parameters(self,):
        bn_params = list(self.block1[1].parameters()) +\
            list(self.block1[4].parameters()) +\
            list(self.block1[7].parameters()) +\
            list(self.block2[1].parameters()) +\
            list(self.block2[4].parameters()) +\
            list(self.block2[7].parameters()) +\
            list(self.block3[1].parameters()) +\
            list(self.block3[4].parameters()) +\
            list(self.classifier[1].parameters())

        other_params = list(self.block1[0].parameters()) +\
            list(self.block1[3].parameters()) +\
            list(self.block1[6].parameters()) +\
            list(self.block2[0].parameters()) +\
            list(self.block2[3].parameters()) +\
            list(self.block2[6].parameters()) +\
            list(self.block3[0].parameters()) +\
            list(self.block3[3].parameters()) +\
            list(self.classifier[0].parameters())

        return bn_params, other_params



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod


class BaseLearner(ABC):
    """
    The BaseLearner class contains all shared methods and attributes
    between the learner classes.
    It is extended by Learner, LearnerRL and LearnerActive that add
    extra abstract methods.

    This class provides abstract methods for:
    - saving the model (save())
    - loading the model (load())
    - optimizing the model (optimize())
    - resetting the model's state if required (reset())
    """

    def __init__(self, lr=0.001, iters=10, batch_size=64, optimizer='sgd', lr_schedule='',
                 backbone='default', network_head='', checkpoint_after_iter=0, checkpoint_load_iter=0,
                 temp_path='', device='cuda', threshold=0.0, scale=1.0):

        self._model = None  # Protected attribute; reference to the model object

        # All parameters below are public attributes with appropriate getters/setters
        # Training parameters
        self.lr = lr
        self.lr_schedule = lr_schedule
        self.iters = iters
        self.batch_size = batch_size
        self.optimizer = optimizer
        self.temp_path = temp_path
        self.checkpoint_after_iter = checkpoint_after_iter
        self.checkpoint_load_iter = checkpoint_load_iter
        self.backbone = backbone
        self.network_head = network_head
        self.device = device

        # Inference parameters
        self.threshold = threshold
        self.scale = scale

    @abstractmethod
    def save(self, path):
        """
        Saves the model in the path provided.

        :param path: Path to save directory
        :type path: str
        :return: Whether save succeeded or not
        :rtype: bool
        """
        pass

    @abstractmethod
    def load(self, path):
        """
        Loads a model from the path provided.

        :param path: Path to saved model
        :type path: str
        :return: Whether load succeeded or not
        :rtype: bool
        """
        pass

    @abstractmethod
    def optimize(self, target_device):
        """
        This method optimizes the model based on the parameters provided.

        :param target_device: the optimization's procedure target device
        :type target_device: str
        :return: Whether optimize succeeded or not
        :rtype: bool
        """
        pass

    @abstractmethod
    def reset(self):
        """
        In the case of stateful models, this method can be used to reset
        the model to its initial state.

        :return: None
        :rtype: None
        """
        pass

    @property
    def model(self):
        """
        Getter of _model field.
        This returns the model.

        :return: the neural network model used by this learner
        :rtype: any
        """
        return self._model

    @model.setter
    def model(self, value):
        """
        Setter of _model field.

        :param value: new model value
        :type value: any
        """
        self._model = value

    @property
    def lr(self):
        """
        Getter of learning rate field.
        This returns the value of the learning rate.

        :return: the value of the learning rate
        :rtype: float
        """
        return self._lr

    @lr.setter
    def lr(self, value):
        """
        Setter for learning rate. This will perform the necessary type checking.

        :param value: new lr value
        :type value: float
        """
        if type(value) != float:
            raise TypeError('lr should be a float')
        else:
            self._lr = value

    @property
    def iters(self):
        """
        Getter of iterations field.
        This returns the number of iterations to be run.

        :return: the value of the iterations
        :rtype: int
        """
        return self._iters

    @iters.setter
    def iters(self, value):
        """
        Setter for iterations. This will perform the necessary type checking.

        :param value: new iters value
        :type value: int
        """
        if type(value) != int:
            raise TypeError('iters should be an int')
        else:
            self._iters = value

    @property
    def batch_size(self):
        """
        Getter of batch size field.
        This returns the size of the batch.

        :return: the value of the batch size
        :rtype: int
        """
        return self._batch_size

    @batch_size.setter
    def batch_size(self, value):
        """
        Setter for batch size. This will perform the necessary type checking.

        :param value: new batch_size value
        :type value: int
        """
        if type(value) != int:
            raise TypeError('batch_size should be an int')
        else:
            self._batch_size = value

    @property
    def optimizer(self):
        """
        Getter of optimizer field.
        This returns the optimizer to be used during training.

        :return: the name of the optimizer
        :rtype: str
        """
        return self._optimizer

    @optimizer.setter
    def optimizer(self, value):
        """
        Setter for optimizer. This will perform the necessary type checking.

        :param value: new optimizer value
        :type value: str
        """
        if type(value) != str:
            raise TypeError('optimizer should be a str')
        else:
            self._optimizer = value

    @property
    def lr_schedule(self):
        """
        Getter of learning rate scheduler field.
        This returns the lr scheduler to be used during training.

        :return: the name of the lr scheduler
        :rtype: str
        """
        return self._lr_schedule

    @lr_schedule.setter
    def lr_schedule(self, value):
        """
        Setter for lr scheduler. This will perform the necessary type checking.

        :param value: new lr_schedule value
        :type value: str
        """
        if type(value) != str:
            raise TypeError('lr_schedule should be a str')
        else:
            self._lr_schedule = value

    @property
    def backbone(self):
        """
        Getter of backbone field.
        This returns the backbone of the model.

        :return: the backbone of the model
        :rtype: str
        """
        return self._backbone

    @backbone.setter
    def backbone(self, value):
        """
        Setter for backbone architecture. This will perform the necessary type checking.

        :param value: new backbone value
        :type value: str
        """
        if type(value) != str:
            raise TypeError('backbone should be a str')
        else:
            self._backbone = value

    @property
    def network_head(self):
        """
        Getter of network head field.
        This returns the network's head to be used.

        :return: the head of the model
        :rtype: str
        """
        return self._network_head

    @network_head.setter
    def network_head(self, value):
        """
        Setter for model's  head. This will perform the necessary type checking.

        :param value: new network_head value
        :type value: str
        """
        if type(value) != str:
            raise TypeError('network_head should be a str')
        else:
            self._network_head = value

    @property
    def checkpoint_after_iter(self):
        """
        Getter of checkpoint_after_iter field.
        If set to 0, no checkpoints will be created.

        :return: the number of iters between checkpoints
        :rtype: int
        """
        return self._checkpoint_after_iter

    @checkpoint_after_iter.setter
    def checkpoint_after_iter(self, value):
        """
        Setter for checkpoint_after_iter. This will perform the necessary type checking.

        :param value: new checkpoint_after_iter value
        :type value: int
        """
        if type(value) != int:
            raise TypeError('checkpoint_after_iter should be an int')
        else:
            self._checkpoint_after_iter = value

    @property
    def checkpoint_load_iter(self):
        """
        Getter of checkpoint_load_iter field.
        This defines what checkpoint to load based on number of iters.
        If set to 0, no checkpoints will be loaded.

        :return: the number of iters
        :rtype: int
        """
        return self._checkpoint_load_iter

    @checkpoint_load_iter.setter
    def checkpoint_load_iter(self, value):
        """
        Setter for checkpoint_load_iter. This will perform the necessary type checking.

        :param value: new checkpoint_load_iter value
        :type value: int
        """
        if type(value) != int:
            raise TypeError('checkpoint_load_iter should be a int')
        else:
            self._checkpoint_load_iter = value

    @property
    def temp_path(self):
        """
        Getter of temp_path field.
        This returns the path of the temporary directory.

        :return: the path of the temporary directory
        :rtype: str
        """
        return self._temp_path

    @temp_path.setter
    def temp_path(self, value):
        """
        Setter for temp_path. This will perform the necessary type checking.

        :param value: new temp_path value
        :type value: str
        """
        # TODO Maybe check for path validity/existence?
        if type(value) != str:
            raise TypeError('temp_path should be a str')
        else:
            self._temp_path = value

    @property
    def device(self):
        """
        Getter of device field.
        This returns the value of the device.

        :return: the value of the device
        :rtype: str
        """
        return self._device

    @device.setter
    def device(self, value):
        """
        Setter for device. This will perform the necessary type checking.

        :param value: new device value
        :type value: str
        """
        if type(value) != str:
            raise TypeError('device should be a str')
        else:
            self._device = value

    @property
    def threshold(self):
        """
        Getter of threshold field. Returns the threshold field value.
        If set to a value different from 0, then allows for using a custom detection/recognition threshold.

        :return: The threshold value
        :rtype: float
        """
        return self._threshold

    @threshold.setter
    def threshold(self, value):
        """
        Setter for threshold. This will perform the necessary type checking.

        :param value: New threshold value
        :type value: float
        """
        if type(value) != float:
            raise TypeError('threshold should be a float')
        else:
            self._threshold = value

    @property
    def scale(self):
        """
        Allows for scaling down the input (values < 1), to accelerate the inference process, or to perform
        analysis on higher resolution than the default (values > 1).

        :return: The scale value
        :rtype: float
        """
        return self._scale

    @scale.setter
    def scale(self, value):
        """
        Setter for scale. This will perform the necessary type checking.

        :param value: New scale value
        :type value: float
        """
        if type(value) != float:
            raise TypeError('scale should be a float')
        else:
            self._scale = value

    @staticmethod
    def get_hyperparameters():
        """
        Static method for obtaining the names of hyperparameters that can be tuned, their type and their possible
        values. This method allows to easily perform hyperparameter tuning.

        :return: Description of hyperparameters.
        :rtype: list[dict[str, any]]
        """
        return None

    @staticmethod
    def get_objective_function():
        """
        Static method for obtaining a mapping from the output of the eval method to a scalar objective that is to be
        minimized during hyperparameter tuning. This method allows to easily perform hyperparameter tuning.

        :return: Objective function that maps the output from the eval_method to an objective scalar value.
        :rtype: callable[any]
        """
        return None


class Learner(BaseLearner):
    """
    The Learner abstract class can be used to implement most perception algorithms.

    It extends the BaseLearner class by adding methods whose implementations
    should be specific for non-RL, non-active algorithms.

    All classes responsible for implementing any perception algorithm should
    inherit and implement the abstract Learner class to ensure that a common
    interface will be provided for training.

    This class extends BaseLearner by adding the following abstract methods for
    non-RL, non-active algorithms:
    - training algorithms (fit())
    - evaluating the performance of a trained model (eval())
    - performing inference (infer())
    """

    def __init__(self, lr=0.001, iters=10, batch_size=64, optimizer='sgd', lr_schedule='',
                 backbone='default', network_head='', checkpoint_after_iter=0, checkpoint_load_iter=0,
                 temp_path='', device='cuda', threshold=0.0, scale=1.0):
        super(Learner, self).__init__(lr=lr, iters=iters, batch_size=batch_size, optimizer=optimizer,
                                      lr_schedule=lr_schedule, backbone=backbone, network_head=network_head,
                                      checkpoint_after_iter=checkpoint_after_iter,
                                      checkpoint_load_iter=checkpoint_load_iter, temp_path=temp_path,
                                      device=device, threshold=threshold, scale=scale)

    @abstractmethod
    def fit(self, dataset, val_dataset=None, logging_path='', silent=True, verbose=True):
        """
        This method is used for training the algorithm on a train dataset and
        validating on a val dataset.

        Can be parameterized based on the learner attributes and custom hyperparameters
        added by the implementation and returns stats regarding training and validation.

        :param dataset: Object that holds the training dataset
        :type dataset: Dataset class type
        :param val_dataset: Object that holds the validation dataset
        :type val_dataset: Dataset class type, optional
        :param verbose: if set to True, enables the maximum logging verbosity (depends on the actual algorithm)
        :type verbose: bool, optional
        :param silent: if set to True, disables printing training progress reports to STDOUT
        :type silent: bool, optional
        :param logging_path: path to save tensorboard log files. If set to None or ‘’, tensorboard logging is disabled
        :type logging_path: str, optional
        :return: Returns stats regarding training and validation
        :rtype: dict
        """
        pass

    @abstractmethod
    def eval(self, dataset):
        """
        This method is used to evaluate the algorithm on a dataset
        and returns stats regarding the evaluation ran.

        :param dataset: Object that holds the dataset to evaluate the algorithm on
        :type dataset: Dataset class type
        :return: Returns stats regarding evaluation
        :rtype: dict
        """
        pass

    @abstractmethod
    def infer(self, batch):
        """
        This method performs inference on the batch provided.

        Can be parameterized based on the learner attributes and custom hyperparameters
        added by the implementation and returns a list of predicted targets.

        :param batch: Object that holds a batch of data to run inference on
        :type batch: Data class type
        :return: A list of predicted targets
        :rtype: list of Target class type objects
        """
        pass


class LearnerRL(BaseLearner):
    """
    The LearnerRL abstract class can be used to implement algorithms trained
    using RL.

    It extends the BaseLearner class by adding methods whose implementations
    should be specific for RL algorithms.

    This class extends BaseLearner by adding the following abstract methods for
    RL algorithms:
    - training algorithms (fit())
    - evaluating the performance of a trained model (eval())
    - performing inference (infer())
    """

    def __init__(self, lr=0.001, iters=10, batch_size=64, optimizer='sgd', lr_schedule='',
                 backbone='default', network_head='', checkpoint_after_iter=0, checkpoint_load_iter=0,
                 temp_path='', device='cuda', threshold=0.0, scale=1.0):
        super(LearnerRL, self).__init__(lr=lr, iters=iters, batch_size=batch_size, optimizer=optimizer,
                                        lr_schedule=lr_schedule, backbone=backbone, network_head=network_head,
                                        checkpoint_after_iter=checkpoint_after_iter,
                                        checkpoint_load_iter=checkpoint_load_iter, temp_path=temp_path,
                                        device=device, threshold=threshold, scale=scale)

    @abstractmethod
    def fit(self, env, val_env=None, logging_path='', silent=True, verbose=True):
        """
        This method is used for training the RL algorithm on a
        training environment and validating on a validation environment.

        Can be parameterized based on the learner attributes and custom hyperparameters
        added by the implementation and returns stats regarding training and validation.

        :param env: Object that holds the training environment
        :type env: BaseEnv class type
        :param val_env: Object that holds the validation environment
        :type val_env: BaseEnv class type
        :param verbose: if set to True, enables the maximum logging verbosity (depends on the actual algorithm)
        :type verbose: bool, optional
        :param silent: if set to True, disables printing training progress reports to STDOUT
        :type silent: bool, optional
        :param logging_path: path to save tensorboard log files. If set to None or ‘’, tensorboard logging is disabled
        :type logging_path: str, optional
        :return: Returns stats regarding training and validation
        :rtype: dict
        """
        pass

    @abstractmethod
    def eval(self, env):
        """
        This method is used to evaluate the algorithm on an environment
        and returns stats regarding the evaluation ran.

        :param env: Object that holds the environment to evaluate the algorithm on
        :type env: BaseEnv class type
        :return: Returns stats regarding evaluation
        :rtype: dict
        """
        pass

    @abstractmethod
    def infer(self, batch):
        """
        This method performs inference on the batch provided.

        Can be parameterized based on the learner attributes and custom hyperparameters
        added by the implementation and returns a list of predicted RL targets.

        :param batch: Object that holds a batch of data to run inference on
        :type batch: Data class type
        :return: A list of predicted RL targets
        :rtype: list of RLTarget class type objects
        """
        pass


class LearnerActive(BaseLearner):
    """
    The LearnerActive abstract class can be used to implement active algorithms.

    It extends the BaseLearner class by adding methods whose implementations
    should be specific for active algorithms.

    This class extends BaseLearner by adding the following abstract methods for
    active algorithms:
    - training algorithms (fit())
    - evaluating the performance of a trained model (eval())
    - performing inference (infer())
    """

    def __init__(self, lr=0.001, iters=10, batch_size=64, optimizer='sgd', lr_schedule='',
                 backbone='default', network_head='', checkpoint_after_iter=0, checkpoint_load_iter=0,
                 temp_path='', device='cuda', threshold=0.0, scale=1.0):
        super(LearnerActive, self).__init__(lr=lr, iters=iters, batch_size=batch_size, optimizer=optimizer,
                                            lr_schedule=lr_schedule, backbone=backbone, network_head=network_head,
                                            checkpoint_after_iter=checkpoint_after_iter,
                                            checkpoint_load_iter=checkpoint_load_iter, temp_path=temp_path,
                                            device=device, threshold=threshold, scale=scale)

    @abstractmethod
    def fit(self, dataset, env, val_dataset=None, val_env=None, logging_path='', silent=True, verbose=True):
        """
        This method is used for training the algorithm on a training
        environment/dataset and validating on a validation environment/dataset.

        Can be parameterized based on the algorithm with the hyperparameters
        argument and returns stats regarding training and validation.

        :param dataset: Object that holds the training dataset
        :type dataset: Dataset class type
        :param env: Object that holds the training environment
        :type env: BaseEnv class type
        :param val_dataset: Object that holds the validation dataset
        :type val_dataset: Dataset class type
        :param val_env: Object that holds the validation environment
        :type val_env: BaseEnv class type
        :param verbose: if set to True, enables the maximum logging verbosity (depends on the actual algorithm)
        :type verbose: bool, optional
        :param silent: if set to True, disables printing training progress reports to STDOUT
        :type silent: bool, optional
        :param logging_path: path to save tensorboard log files. If set to None or ‘’, tensorboard logging is disabled
        :type logging_path: str, optional
        :return: Returns stats regarding training and validation
        :rtype: dict
        """
        pass

    @abstractmethod
    def eval(self, dataset, env):
        """
        This method is used to evaluate the algorithm on an environment/dataset
        and returns stats regarding the evaluation ran.

        :param dataset: Object that holds the dataset to evaluate the algorithm on
        :type dataset: Dataset class type
        :param env: Object that holds the environment to evaluate the algorithm on
        :type env: BaseEnv class type
        :return: Returns stats regarding evaluation
        :rtype: dict
        """
        pass

    @abstractmethod
    def infer(self, batch):
        """
        This method performs inference on the batch provided.

        Can be parameterized based on the learner attributes and custom hyperparameters
        added by the implementation and returns a list of predicted targets.

        :param batch: Object that holds a batch of data to run inference on
        :type batch: Data class type
        :return: A list of predicted targets
        :rtype: list of Target class type objects
        """
        pass



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# cifar
from .cifar_allcnn import AllCNN as CifarAllCNN
from .cifar_densenet import DenseNet121 as CifarDenseNet121
from .cifar_densenet import DenseNet169 as CifarDenseNet169
from .cifar_densenet import DenseNet201 as CifarDenseNet201
from .cifar_densenet import DenseNet161 as CifarDenseNet161
from .cifar_resnet import ResNet18 as CifarResNet18
from .cifar_resnet import ResNet34 as CifarResNet34
from .cifar_resnet import ResNet50 as CifarResNet50
from .cifar_resnet import ResNet101 as CifarResNet101
from .cifar_resnet import ResNet152 as CifarResNet152
from .cifar_vgg import VGG11 as CifarVGG11
from .cifar_vgg import VGG13 as CifarVGG13
from .cifar_vgg import VGG16 as CifarVGG16
from .cifar_vgg import VGG19 as CifarVGG19

# imagenet
from .imagenet_vgg import VGG11, VGG11_BN, VGG13, VGG13_BN, VGG16, VGG16_BN, VGG19, VGG19_BN
from .imagenet_resnet import ResNet18, ResNet34, ResNet50, ResNet101, ResNet152
from .imagenet_resnet import ResNext50_32x4d, ResNext101_32x8d, WideResNet50_2, WideResNet101_2
from .imagenet_densenet import DenseNet121, DenseNet161, DenseNet169, DenseNet201

models = {'cifar_allcnn': CifarAllCNN,
          'cifar_densenet121': CifarDenseNet121,
          'cifar_densenet169': CifarDenseNet169,
          'cifar_densenet201': CifarDenseNet201,
          'cifar_densenet161': CifarDenseNet161,
          'cifar_resnet18': CifarResNet18,
          'cifar_resnet34': CifarResNet34,
          'cifar_resnet50': CifarResNet50,
          'cifar_resnet101': CifarResNet101,
          'cifar_resnet152': CifarResNet152,
          'cifar_vgg11': CifarVGG11,
          'cifar_vgg13': CifarVGG13,
          'cifar_vgg16': CifarVGG16,
          'cifar_vgg19': CifarVGG19,
          'imagenet_vgg11': VGG11,
          'imagenet_vgg11_bn': VGG11_BN,
          'imagenet_vgg13': VGG13,
          'imagenet_vgg13_bn': VGG13_BN,
          'imagenet_vgg16': VGG16,
          'imagenet_vgg16_bn': VGG16_BN,
          'imagenet_vgg19': VGG19,
          'imagenet_vgg19_bn': VGG19_BN,
          'imagenet_resnet18': ResNet18,
          'imagenet_resnet34': ResNet34,
          'imagenet_resnet50': ResNet50,
          'imagenet_resnet101': ResNet101,
          'imagenet_resnet152': ResNet152,
          'imagenet_resnext50_32x4d': ResNext50_32x4d,
          'imagenet_resnext101_32x8d': ResNext101_32x8d,
          'imagenet_wide_resnet50_2': WideResNet50_2,
          'imagenet_wide_resnet101_2': WideResNet101_2,
          'imagenet_densenet121': DenseNet121,
          'imagenet_densenet161': DenseNet161,
          'imagenet_densenet169': DenseNet169,
          'imagenet_densenet201': DenseNet201,
          }



# MIT License
#
# Copyright (c) 2017 liukuang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from . import model_utils


class Bottleneck(nn.Module):
    def __init__(self, in_planes, growth_rate):
        super(Bottleneck, self).__init__()
        self.bn1 = nn.BatchNorm2d(in_planes)
        self.conv1 = nn.Conv2d(in_planes, 4 * growth_rate, kernel_size=1, bias=False)
        self.bn2 = nn.BatchNorm2d(4 * growth_rate)
        self.conv2 = nn.Conv2d(4 * growth_rate, growth_rate, kernel_size=3, padding=1, bias=False)

    def forward(self, x):
        out = self.conv1(F.relu(self.bn1(x)))
        out = self.conv2(F.relu(self.bn2(out)))
        out = torch.cat([out, x], 1)
        return out

    def get_parameters(self,):
        bn_params = list(self.bn1.parameters()) + list(self.bn2.parameters())
        other_params = list(self.conv1.parameters()) + list(self.conv2.parameters())

        return bn_params, other_params


class Transition(nn.Module):
    def __init__(self, in_planes, out_planes):
        super(Transition, self).__init__()
        self.bn = nn.BatchNorm2d(in_planes)
        self.conv = nn.Conv2d(in_planes, out_planes, kernel_size=1, bias=False)

    def forward(self, x):
        out = self.conv(F.relu(self.bn(x)))
        out = F.avg_pool2d(out, 2)
        return out

    def get_parameters(self,):
        return list(self.bn.parameters()), list(self.conv.parameters())


class DenseNet(nn.Module):
    def __init__(self, block, nblocks, growth_rate=12, reduction=0.5, num_classes=10):
        super(DenseNet, self).__init__()
        self.growth_rate = growth_rate

        num_planes = 2 * growth_rate
        self.conv1 = nn.Conv2d(3, num_planes, kernel_size=3, padding=1, bias=False)

        dense_layers = []
        dense_layers += self._make_dense_layers(block, num_planes, nblocks[0])
        num_planes += nblocks[0] * growth_rate
        out_planes = int(math.floor(num_planes*reduction))
        dense_layers.append(Transition(num_planes, out_planes))
        num_planes = out_planes

        dense_layers += self._make_dense_layers(block, num_planes, nblocks[1])
        num_planes += nblocks[1] * growth_rate
        out_planes = int(math.floor(num_planes * reduction))
        dense_layers.append(Transition(num_planes, out_planes))
        num_planes = out_planes

        dense_layers += self._make_dense_layers(block, num_planes, nblocks[2])
        num_planes += nblocks[2] * growth_rate
        out_planes = int(math.floor(num_planes * reduction))
        dense_layers.append(Transition(num_planes, out_planes))
        num_planes = out_planes

        dense_layers += self._make_dense_layers(block, num_planes, nblocks[3])
        num_planes += nblocks[3] * growth_rate

        self.dense_layers = nn.ModuleList(dense_layers)

        self.bn = nn.BatchNorm2d(num_planes)
        self.linear = nn.Linear(num_planes, num_classes)

        self.initialize()

    def _make_dense_layers(self, block, in_planes, nblock):
        layers = []
        for i in range(nblock):
            layers.append(block(in_planes, self.growth_rate))
            in_planes += self.growth_rate

        return layers

    def forward(self, x):
        out = self.conv1(x)

        for layer in self.dense_layers:
            out = layer(out)

        out = F.relu(self.bn(out))
        out = out.mean(-1).mean(-1)
        out = self.linear(out)

        return out

    def initialize(self,):
        for layer in self.modules():
            if isinstance(layer, nn.Conv2d):
                nn.init.kaiming_normal_(layer.weight)
            elif isinstance(layer, nn.BatchNorm2d):
                nn.init.constant_(layer.weight, 1)
                nn.init.constant_(layer.bias, 0.0)
            elif isinstance(layer, nn.Linear):
                nn.init.kaiming_normal_(layer.weight)
                nn.init.constant_(layer.bias, 0.0)

    def get_parameters(self,):
        bn_params, other_params = [], []

        other_params.extend(list(self.conv1.parameters()))

        for layer in self.dense_layers:
            bn, other = layer.get_parameters()
            bn_params.extend(bn)
            other_params.extend(other)

        bn_params.extend(list(self.bn.parameters()))
        other_params.extend(list(self.linear.parameters()))

        return bn_params, other_params


def load_pretrained_weights(model, model_name, n_class, pretrained):
    if pretrained == 'with_classifier':
        if n_class == 10:
            model_name = 'cifar10_{}'.format(model_name)
        elif n_class == 100:
            model_name = 'cifar100_{}'.format(model_name)
        else:
            model_name = ''

        state_dict = model_utils.get_cifar_pretrained_weights(model_name)

        if state_dict is not None:
            model.load_state_dict(state_dict)

    elif pretrained == 'without_classifier':
        model_name = 'cifar100_{}'.format(model_name)
        state_dict = model_utils.get_cifar_pretrained_weights(model_name)
        if state_dict is not None:
            cur_state_dict = model.state_dict()
            state_dict['linear.weight'] = cur_state_dict['linear.weight']
            state_dict['linear.bias'] = cur_state_dict['linear.bias']
            model.load_state_dict(state_dict)


class DenseNet121(DenseNet):
    def __init__(self, n_class, pretrained):
        super(DenseNet121, self).__init__(Bottleneck, [6, 12, 24, 16], growth_rate=32, num_classes=n_class)
        load_pretrained_weights(self, 'densenet121', n_class, pretrained)


class DenseNet169(DenseNet):
    def __init__(self, n_class, pretrained):
        super(DenseNet169, self).__init__(Bottleneck, [6, 12, 32, 32], growth_rate=32, num_classes=n_class)
        load_pretrained_weights(self, 'densenet169', n_class, pretrained)


class DenseNet201(DenseNet):
    def __init__(self, n_class, pretrained):
        super(DenseNet201, self).__init__(Bottleneck, [6, 12, 48, 32], growth_rate=32, num_classes=n_class)
        load_pretrained_weights(self, 'densenet201', n_class, pretrained)


class DenseNet161(DenseNet):
    def __init__(self, n_class, pretrained):
        super(DenseNet161, self).__init__(Bottleneck, [6, 12, 36, 24], growth_rate=48, num_classes=n_class)
        load_pretrained_weights(self, 'densenet161', n_class, pretrained)



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch


class DataWrapper:
    def __init__(self, opendr_dataset):
        self.dataset = opendr_dataset

    def __len__(self,):
        return len(self.dataset)

    def __getitem__(self, i):
        x, y = self.dataset.__getitem__(i)
        # change from rows x cols x channels to channels x rows x cols
        x = x.convert("channels_first", "rgb")
        return torch.from_numpy(x).float(), torch.tensor([y.data, ]).long()



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# general imports
import numpy as np
import torch
from torch.utils.data import DataLoader
import tempfile
import random
import string
import time
import os
import json
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm
from urllib.request import urlretrieve

# OpenDR engine imports
from opendr.engine.learners import Learner
from opendr.engine.data import Image
from opendr.engine.datasets import DatasetIterator
from opendr.engine.target import Category
from opendr.engine.constants import OPENDR_SERVER_URL

# OpenDR multilinear_compressive_learning imports
from opendr.perception.compressive_learning.multilinear_compressive_learning.algorithm import (
    trainers,
    CompressiveLearner,
    get_builtin_backbones
)
from opendr.perception.compressive_learning.multilinear_compressive_learning.algorithm.data import DataWrapper
from opendr.perception.compressive_learning.multilinear_compressive_learning.algorithm.trainers import (
    get_cosine_lr_scheduler,
    get_multiplicative_lr_scheduler
)

__all__ = ['get_cosine_lr_scheduler',
           'get_multiplicative_lr_scheduler',
           'get_builtin_backbones']

# constants
PRETRAINED_COMPRESSED_SHAPE = ((20, 19, 2),
                               (28, 27, 1),
                               (14, 11, 2),
                               (18, 17, 1),
                               (9, 6, 1),
                               (6, 9, 1))


class MultilinearCompressiveLearner(Learner):
    def __init__(self,
                 input_shape,
                 compressed_shape,
                 backbone,
                 n_class,
                 pretrained_backbone='',
                 init_backbone=True,
                 lr_scheduler=get_cosine_lr_scheduler(1e-3, 1e-5),
                 optimizer='adam',
                 weight_decay=1e-4,
                 n_init_iters=100,
                 iters=300,
                 batch_size=32,
                 checkpoint_after_iter=1,
                 checkpoint_load_iter=0,
                 temp_path='',
                 device='cpu',
                 test_mode=False,
                 ):

        super(MultilinearCompressiveLearner, self).__init__(batch_size=batch_size,
                                                            checkpoint_after_iter=checkpoint_after_iter,
                                                            checkpoint_load_iter=checkpoint_load_iter,
                                                            temp_path=temp_path,
                                                            device=device)

        assert pretrained_backbone in ['', 'with_classifier', 'without_classifier'],\
            'Option "pretrained_backbone" must be empty, "with_classifier" or "without_classifier". ' +\
            'Provided value: {}'.format(pretrained_backbone)

        if pretrained_backbone == 'with_classifier':
            if isinstance(backbone, str) and backbone.startswith('cifar'):
                assert n_class in [10, 100],\
                    'For `backbone`="{}" and `pretrained_backbone`="with_classifier", '.format(backbone) +\
                    '`n_class` must be 10 or 100\n' +\
                    'Provided `n_class`={}'.format(n_class)
            elif isinstance(backbone, str) and backbone.startswith('imagenet'):
                assert n_class == 1000,\
                    'For `backbone`="{}" and `pretrained_backbone`="with_classifier", '.format(backbone) +\
                    '`n_class` must be 1000\n' +\
                    'Provided `n_class`={}'.format(n_class)

        assert checkpoint_load_iter in [-1, 0],\
            'check_point_load_iter must be -1 or 0, with 0 indicating training from scratch,' +\
            '-1 indicating training from latest checkpoint if temp_path is given'

        assert optimizer in ['adam', 'sgd'],\
            'given optimizer "{}" is not supported, please select set optimizer to "adam" or "sgd"'.format(optimizer)

        self.model = CompressiveLearner(input_shape, compressed_shape, n_class, backbone, pretrained_backbone)
        self.input_shape = input_shape
        self.compressed_shape = compressed_shape
        self.backbone_classifier = backbone
        self.init_backbone = init_backbone
        self.n_class = n_class
        self.lr_scheduler = lr_scheduler
        self.n_init_epoch = n_init_iters
        self.n_epoch = iters
        self.batch_size = batch_size
        self.checkpoint_freq = checkpoint_after_iter
        self.epoch_idx = checkpoint_load_iter
        self.optimizer = optimizer
        self.weight_decay = weight_decay
        self.device = device
        self.test_mode = test_mode
        self.temp_path = temp_path

    def _prepare_temp_dir(self,):
        if self.temp_path == '':
            # if temp dir not provided, create one under default system temp dir
            random_name = ''.join([random.choice(string.ascii_lowercase) for _ in range(128)]) + str(time.time())
            self.temp_dir_obj = tempfile.TemporaryDirectory(prefix=random_name)
            temp_dir = self.temp_dir_obj.name
        else:
            if not os.path.exists(self.temp_path):
                os.makedirs(self.temp_path)
            self.temp_dir_obj = None
            temp_dir = self.temp_path

        self.classifier_temp_dir = os.path.join(temp_dir, 'classifier')
        self.initializer_temp_dir = os.path.join(temp_dir, 'initializer')
        self.model_temp_dir = os.path.join(temp_dir, 'main_model')

        if not os.path.exists(self.classifier_temp_dir):
            os.mkdir(self.classifier_temp_dir)

        if not os.path.exists(self.initializer_temp_dir):
            os.mkdir(self.initializer_temp_dir)

        if not os.path.exists(self.model_temp_dir):
            os.mkdir(self.model_temp_dir)

    def _validate_dataset(self, dataset):
        """
        This internal function is used to perform basic validation of the data dimensions
        """

        if dataset is None:
            return

        if not isinstance(dataset, DatasetIterator):
            msg = 'Dataset must be an instance of `engine.datasets.DatasetIterator` class\n' +\
                  'Received an instance of type: {}'.format(type(dataset))
            raise TypeError(msg)

        x, y = dataset.__getitem__(0)

        if not isinstance(x, Image):
            msg = 'The 1st element returned by __getitem__ must be an instance of `engine.data.Image` class\n' +\
                  'Received an instance of type: {}'.format(type(x))
            raise TypeError(msg)

        if not isinstance(y, Category):
            msg = 'The 2nd element returned by __getitem__ must be an instance of `engine.target.Cateogry` class\n' +\
                  'Received an instance of type: {}'.format(type(y))
            raise TypeError(msg)

        x = x.convert("channels_last", "rgb")
        if not x.shape == tuple(self.input_shape):
            msg = 'Given input_shape does not match dimensions of data produced by dataset\n' +\
                  'input_shape: {}, data shape: {}'.format(tuple(self.input_shape), x.shape)
            raise ValueError(msg)

    def fit(self, train_set, val_set=None, test_set=None, logging_path='', silent=False, verbose=True):
        """
        Method to train the multilinear compressive learning model

        :param train_set: object that holds training samples
        :type train_set: engine.datasets.DatasetIterator
                         with __getitem__ producing (engine.data.Image, engine.target.Category)
        :param val_set: object that holds the validation samples, default to None
                        if available, `val_set` is used to select the best checkpoint for final model
        :type val_set: engine.datasets.DatasetIterator
                       with __getitem__ producing (engine.data.Image, engine.target.Category)
        :param test_set: object that holds the test samples, default to None
        :type test_set: engine.datasets.DatasetIterator
                        with __getitem__ producing (engine.data.Image, engine.target.Category)
        :param logging_path: path to save tensoboard data, default to ""
        :type logging_path: string
        :param silent: disable performance printing, default to False
        :type silent: bool
        :param verbose: enable the progress bar of each epoch, default to True
        :type verbose: bool

        :return: returns the performance curves when training
                 backbone classifier, sensing and synthesis components and compressive learning model
        :rtype: list(dict)
        """

        self._validate_dataset(train_set)
        self._validate_dataset(val_set)
        self._validate_dataset(test_set)
        self._prepare_temp_dir()

        train_loader = DataLoader(DataWrapper(train_set),
                                  batch_size=self.batch_size,
                                  pin_memory='cuda' in self.device,
                                  shuffle=True)

        if val_set is None:
            val_loader = None
        else:
            val_loader = DataLoader(DataWrapper(val_set),
                                    batch_size=self.batch_size,
                                    pin_memory='cuda' in self.device,
                                    shuffle=False)
        if test_set is None:
            test_loader = None
        else:
            test_loader = DataLoader(DataWrapper(test_set),
                                     batch_size=self.batch_size,
                                     pin_memory='cuda' in self.device,
                                     shuffle=False)

        if self.test_mode and not silent:
            print('\nWARNING: training model under test mode\n')

        if logging_path != '':
            if not os.path.exists(logging_path):
                os.makedirs(logging_path)

            tensorboard_logger = SummaryWriter(logging_path)
        else:
            tensorboard_logger = None

        # train the backbone classifier if init_backbone is True
        if self.init_backbone:
            backbone_performance = self._fit_backbone(train_loader,
                                                      val_loader,
                                                      test_loader,
                                                      tensorboard_logger,
                                                      silent,
                                                      verbose)
        else:
            backbone_performance = {}

        # train the sensing and feature synthesis modules
        init_performance = self._fit_sensing_and_synthesis(train_loader,
                                                           val_loader,
                                                           test_loader,
                                                           tensorboard_logger,
                                                           silent,
                                                           verbose)

        # train the whole model
        compressive_learning_performance = self._fit_system(train_loader,
                                                            val_loader,
                                                            test_loader,
                                                            tensorboard_logger,
                                                            silent,
                                                            verbose)

        if tensorboard_logger is not None:
            tensorboard_logger.close()

        if self.temp_dir_obj is not None:
            self.temp_dir_obj.cleanup()

        performance = {'backbone_performance': backbone_performance,
                       'initialization_performance': init_performance,
                       'compressive_learning_performance': compressive_learning_performance}

        return performance

    def eval(self, dataset, silent=False, verbose=True):
        """
        This method is used to evaluate the performance of a given set of data

        :param dataset: object that holds the set of samples to evaluate
        :type dataset: engine.datasets.DatasetIterator
                       with __getitem__ producing (engine.data.Image, engine.target.Category)

        :return: a dictionary that contains `cross_entropy` and `acc` as keys
        :rtype: dict
        """

        self._validate_dataset(dataset)
        loader = DataLoader(DataWrapper(dataset),
                            batch_size=self.batch_size,
                            pin_memory='cuda' in self.device,
                            shuffle=False)

        device = torch.device(self.device)

        self.model.to(device)
        self.model.eval()

        L = torch.nn.CrossEntropyLoss()
        n_correct = 0
        n_sample = 0
        loss = 0

        if verbose:
            loader = tqdm(loader, desc='#Evaluating ', ncols=80, ascii=True)

        with torch.no_grad():
            for inputs, targets in loader:
                inputs = inputs.to(device)
                targets = targets.to(device).long().flatten()

                predictions = self.model(inputs)
                n_sample += inputs.size(0)
                loss += L(predictions, targets).item()
                n_correct += (predictions.argmax(dim=-1) == targets).sum().item()

        metrics = {'cross_entropy': loss / n_sample,
                   'acc': n_correct / n_sample}

        if not silent:
            print('cross_entropy: {:.6f}, acc: {:.4f}'.format(metrics['cross_entropy'], metrics['acc']))

        return metrics

    def infer(self, img):
        """
        This method is used to generate class prediction given an image

        :param img: image to generate class prediction
        :type img: engine.data.Image

        :return: predicted label
        :rtype: engine.target.Category

        """

        if not isinstance(img, Image):
            msg = 'Input to `infer()` must be an instance of engine.data.Image\n' +\
                  'Received an instance of type: {}'.format(type(img))
            raise TypeError(msg)

        self.model.to(torch.device(self.device))
        self.model.eval()

        img = img.convert("channels_last", "rgb")

        if not tuple(img.shape) == tuple(self.input_shape):
            msg = 'Dimensions of the given input "{}"'.format(img.shape) +\
                  'do not match input dimensions "{}" of the model'.format(tuple(self.input_shape))
            raise ValueError(msg)

        img = np.expand_dims(img.transpose(2, 0, 1), 0)

        with torch.no_grad():
            tensor_img = torch.tensor(img, device=torch.device(self.device)).float()
            prob_prediction = torch.nn.functional.softmax(self.model(tensor_img).flatten(), dim=0)
            class_prediction = prob_prediction.argmax(dim=-1).cpu().item()
            prediction = Category(class_prediction, confidence=prob_prediction[class_prediction].cpu().item())

        return prediction

    def get_sensing_parameters(self):
        """
        This method is used to get the sensing parameters

        :return: parameters of sensing operators
        :rtype: list of numpy array

        """

        params = [x.detach().cpu().numpy() for x in list(self.model.sense_synth_module.synthesis_module.parameters())]
        # reverse the order from pytorch order to normal image order
        if len(params) == 3:
            params = [params[1], params[2], params[0]]

        return params

    def infer_from_compressed_measurement(self, measurement):
        """
        This method is used to generate class prediction given the compressed measurement

        :param measurement: compressed measurement to generate class prediction
        :type measurement: engine.data.Image

        :return: predicted label
        :rtype: engine.target.Category

        """

        if not isinstance(measurement, Image):
            msg = 'Input to `infer_from_compressed_measurement()` must be an instance of engine.data.Image\n' +\
                  'Received an instance of type: {}'.format(type(measurement))
            raise TypeError(msg)

        self.model.to(torch.device(self.device))
        self.model.eval()

        measurement = measurement.convert("channels_last")

        if not tuple(measurement.shape) == tuple(self.compressed_shape):
            msg = 'Dimensions of the given compressed measurement "{}"'.format(measurement.shape) +\
                  'do not match `compressed_shape` "{}" of the model'.format(tuple(self.compressed_shape))
            raise ValueError(msg)

        measurement = np.expand_dims(measurement.transpose(2, 0, 1), 0)

        with torch.no_grad():
            tensor_measurement = torch.tensor(measurement, device=torch.device(self.device)).float()
            prob_prediction = self.model.infer_from_measurement(tensor_measurement).flatten()
            prob_prediction = torch.nn.functional.softmax(prob_prediction, dim=0)
            class_prediction = prob_prediction.argmax(dim=-1).cpu().item()
            prediction = Category(class_prediction, confidence=prob_prediction[class_prediction].cpu().item())

        return prediction

    def save(self, path, verbose=True):
        """
        This function is used to save the current model given a directory path. Metadata and model weights
        are saved under path, i.e., `path/metadata.json` and path/model_weights.pt` will be created

        :param path: path to the directory that the model will be saved
        :type path: str
        :param verbose: whether to print acknowledge message when saving is successful, default to True
        :type verbose: bool

        """

        if not os.path.exists(path):
            os.makedirs(path)

        if isinstance(self.backbone_classifier, str):
            backbone = self.backbone_classifier
        else:
            backbone = type(self.backbone_classifier).__name__

        model_weight_file = os.path.join(path, 'model_weights.pt')
        metadata_file = os.path.join(path, 'metadata.json')

        metadata = {'framework': 'pytorch',
                    'model_paths': model_weight_file,
                    'format': 'pt',
                    'backbone': backbone,
                    'input_shape': self.input_shape,
                    'compressed_shape': self.compressed_shape,
                    'has_data': False,
                    'inference_params': {},
                    'optimized': False,
                    'optimimizer_info': {}
                    }

        try:
            torch.save(self.model.cpu().state_dict(), model_weight_file)
            if verbose:
                print('Model weights saved to {}'.format(model_weight_file))
        except Exception as error:
            raise error

        try:
            fid = open(metadata_file, 'w')
            json.dump(metadata, fid)
            fid.close()

            if verbose:
                print('Model metadata saved to {}'.format(metadata_file))

        except Exception as error:
            raise error

    def load(self, path, verbose=True):
        """
        This function is used to load a pretrained model that has been saved with .save(), given the path
        to the directory. `path/metadata.json` and `path/model_weights.pt` should exist

        :param path: path to the saved location
        :type path: str
        :param verbose: whether to print acknowledge message when loading is successful, defaul to True
        :type verbose: bool

        """

        if not os.path.exists(path):
            raise FileNotFoundError('Directory "{}" does not exist'.format(path))

        if not os.path.isdir(path):
            raise ValueError('Given path "{}" is not a directory'.format(path))

        metadata_file = os.path.join(path, 'metadata.json')
        model_weight_file = os.path.join(path, 'model_weights.pt')

        assert os.path.exists(metadata_file),\
            'Metadata file ("metadata.json") does not exist under the given path "{}"'.format(path)

        assert os.path.exists(model_weight_file),\
            'Model weights ("model_weights.pt") does not exist under the given path "{}"'.format(path)

        fid = open(metadata_file, 'r')
        metadata = json.load(fid)
        fid.close()

        assert tuple(metadata['input_shape']) == tuple(self.input_shape),\
            'Input dimensions "{}" of saved model '.format(metadata['input_shape']) +\
            'do not match input dimensions "{}" of current model instance'.format(self.input_shape)

        assert tuple(metadata['compressed_shape']) == tuple(self.compressed_shape),\
            'Compressed dimensions "{}" of saved model '.format(metadata['compressed_shape']) +\
            'do not match comressed dimensions "{}" of current model instance'.format(self.compressed_shape)

        self.model.cpu()
        self.model.load_state_dict(torch.load(model_weight_file, map_location=torch.device('cpu')))

        if verbose:
            print('Pretrained model is loaded successfully')

    def _fit_backbone(self, train_loader, val_loader, test_loader, tensorboard_logger, silent, verbose):
        """
        This is the internal function used to train the backbone classifier
        """

        if not silent:
            print('\n======= Training backbone classifier =======\n')
        trainer = trainers.ClassifierTrainer(n_epoch=self.n_epoch,
                                             epoch_idx=self.epoch_idx,
                                             lr_scheduler=self.lr_scheduler,
                                             optimizer=self.optimizer,
                                             weight_decay=self.weight_decay,
                                             temp_dir=self.classifier_temp_dir,
                                             checkpoint_freq=self.checkpoint_freq,
                                             print_freq=int(not silent),
                                             use_progress_bar=verbose,
                                             test_mode=self.test_mode)

        device = torch.device(self.device)
        backbone_performance = trainer.fit(self.model.backbone,
                                           train_loader,
                                           val_loader,
                                           test_loader,
                                           device,
                                           tensorboard_logger,
                                           logger_prefix='backbone')

        return backbone_performance

    def _fit_sensing_and_synthesis(self, train_loader, val_loader, test_loader, tensorboard_logger, silent, verbose):
        """
        This is the internal function used to train the sensing and feature synthesis components
        """

        if not silent:
            print('\n======= Training the sensing and feature synthesis modules =======\n')
        trainer = trainers.AutoRegressionTrainer(n_epoch=self.n_init_epoch,
                                                 epoch_idx=self.epoch_idx,
                                                 lr_scheduler=self.lr_scheduler,
                                                 optimizer=self.optimizer,
                                                 weight_decay=self.weight_decay,
                                                 temp_dir=self.initializer_temp_dir,
                                                 checkpoint_freq=self.checkpoint_freq,
                                                 print_freq=int(not silent),
                                                 use_progress_bar=verbose,
                                                 test_mode=self.test_mode)

        device = torch.device(self.device)
        init_performance = trainer.fit(self.model.sense_synth_module,
                                       train_loader,
                                       val_loader,
                                       test_loader,
                                       device,
                                       tensorboard_logger,
                                       logger_prefix='sensing_and_synthesis')

        return init_performance

    def _fit_system(self, train_loader, val_loader, test_loader, tensorboard_logger, silent, verbose):
        """
        This is the internal function used to train the all components of the multilinear compressive
        learning model
        """

        if not silent:
            print('\n======= Training all modules =======\n')
        trainer = trainers.ClassifierTrainer(n_epoch=self.n_epoch,
                                             epoch_idx=self.epoch_idx,
                                             lr_scheduler=self.lr_scheduler,
                                             optimizer=self.optimizer,
                                             weight_decay=self.weight_decay,
                                             temp_dir=self.model_temp_dir,
                                             checkpoint_freq=self.checkpoint_freq,
                                             print_freq=int(not silent),
                                             use_progress_bar=verbose,
                                             test_mode=self.test_mode)

        device = torch.device(self.device)
        backbone_performance = trainer.fit(self.model,
                                           train_loader,
                                           val_loader,
                                           test_loader,
                                           device,
                                           tensorboard_logger,
                                           logger_prefix='main_model')

        return backbone_performance

    def download(self, path):
        """
        This function is used to download a pretrained model given the current model specification
        Calling load(path) after this function will load the downloaded model weights

        :param path: path to the saved location. Under this path `model_weights.pt` and `metadata.json`
                     will be downloaded so different paths for different models should be given to avoid
                     overwriting previously downloaded model
        :type path: str
        """

        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)

        if isinstance(self.backbone_classifier, str) and self.backbone_classifier == 'cifar_allcnn':
            assert self.n_class in [10, 100],\
                'Only support pretrained CIFAR10 or CIFAR100 model.\n' +\
                'Current model specification has {} classes'.format(self.n_class)

            compressed_shape = tuple(self.compressed_shape)
            assert compressed_shape in PRETRAINED_COMPRESSED_SHAPE,\
                'Current `compressed_shape` does not match supported shapes. Supported shapes includes\n' +\
                '\n'.join(PRETRAINED_COMPRESSED_SHAPE)

            server_url = os.path.join(OPENDR_SERVER_URL,
                                      'perception',
                                      'compressive_learning',
                                      'multilinear_compressive_learning')

            dataset = 'cifar10' if self.n_class == 10 else 'cifar100'
            backbone = 'allcnn'
            model_name = '{}_{}_{}_{}_{}'.format(dataset,
                                                 backbone,
                                                 self.compressed_shape[0],
                                                 self.compressed_shape[1],
                                                 self.compressed_shape[2])

            metadata_url = os.path.join(server_url, '{}.json'.format(model_name))
            metadata_file = os.path.join(path, 'metadata.json')
            urlretrieve(metadata_url, metadata_file)

            weights_url = os.path.join(server_url, '{}.pt'.format(model_name))
            weights_file = os.path.join(path, 'model_weights.pt')
            urlretrieve(weights_url, weights_file)
            print('Pretrained model downloaded to the following directory\n{}'.format(path))
        else:
            raise UserWarning('Only pretrained model for built-in backbone can be downloaded')

    def optimize(self):
        pass

    def reset(self):
        pass
